import dataclasses
import itertools
import re
import sys
import tempfile
import uuid
from collections import UserList
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Union

import questionary
import yaml
from questionary import Choice
from rich import print
from rich.markdown import Markdown
from rich.panel import Panel

from cognite_toolkit._cdf_tk.builders import create_builder
from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.constants import BUILD_ENVIRONMENT_FILE, ENV_VAR_PATTERN
from cognite_toolkit._cdf_tk.cruds import (
    ExtractionPipelineConfigCRUD,
    FunctionCRUD,
    GraphQLCRUD,
    GroupAllScopedCRUD,
    HostedExtractorDestinationCRUD,
    HostedExtractorSourceCRUD,
    ResourceCRUD,
    StreamlitCRUD,
)
from cognite_toolkit._cdf_tk.cruds._base_cruds import (
    T_ID,
    T_ResourceRequest,
    T_ResourceRequestList,
    T_ResourceResponse,
    T_ResourceResponseList,
)
from cognite_toolkit._cdf_tk.data_classes import (
    BuildEnvironment,
    BuildVariable,
    BuildVariables,
    BuiltFullResourceList,
    BuiltModuleList,
    BuiltResourceFull,
    DeployResults,
    ModuleDirectories,
    ResourceDeployResult,
    YAMLComments,
)
from cognite_toolkit._cdf_tk.exceptions import ToolkitError, ToolkitMissingResourceError, ToolkitValueError
from cognite_toolkit._cdf_tk.tk_warnings import LowSeverityWarning, MediumSeverityWarning
from cognite_toolkit._cdf_tk.utils import (
    YAMLComment,
    YAMLWithComments,
    read_yaml_content,
    read_yaml_file,
    safe_read,
)
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables
from cognite_toolkit._cdf_tk.utils.file import safe_rmtree, yaml_safe_dump
from cognite_toolkit._cdf_tk.utils.modules import (
    is_module_path,
    module_directory_from_path,
    parse_user_selected_modules,
)

from ._base import ToolkitCommand
from .build_cmd import BuildCommand
from .clean import CleanCommand

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

_VARIABLE_PATTERN = re.compile(r"\{\{(.+?)\}\}")
# The encoding and newline characters to use when writing files
# These are hardcoded to ensure that running the pull command on different platforms
# will produce the same output. The motivation is when having local sources in
# version control, the diff will be easier to read.
ENCODING = "utf-8"
NEWLINE = "\n"


@dataclass
class Variable:
    placeholder: str | None = None
    name: str | None = None
    source_value: str | None = None


@dataclass
class ResourceProperty:
    """This represents a single property in a CDF resource file.

    Args:
        key_path: The path to the property in the resource file.
        build_value: The value of the property in the local resource file build file.
        cdf_value: The value of the property in the CDF resource.
        variables: A list of variables that are used in the property value.
    """

    key_path: tuple[str | int, ...]
    build_value: float | int | str | bool | None = None
    cdf_value: float | int | str | bool | None = None
    variables: list[Variable] = field(default_factory=list)

    @property
    def value(self) -> float | int | str | bool | None:
        if self.has_variables:
            return self.variables[0].source_value
        return self.cdf_value or self.build_value

    @property
    def has_variables(self) -> bool:
        return bool(self.variables)

    @property
    def is_changed(self) -> bool:
        return (
            self.build_value != self.cdf_value
            and self.build_value is not None
            and self.cdf_value is not None
            and not self.has_variables
        )

    @property
    def is_added(self) -> bool:
        return self.build_value is None and self.cdf_value is not None

    @property
    def is_cannot_change(self) -> bool:
        return (
            self.build_value != self.cdf_value
            and not self.has_variables
            and self.build_value is not None
            and self.cdf_value is not None
        )

    def __str__(self) -> str:
        key_str = ".".join(map(str, self.key_path))
        if self.is_added:
            return f"ADDED: '{key_str}: {self.cdf_value}'"
        elif self.is_changed:
            return f"CHANGED: '{key_str}: {self.build_value} -> {self.cdf_value}'"
        elif self.is_cannot_change:
            return f"CANNOT CHANGE (contains variables): '{key_str}: {self.build_value} -> {self.cdf_value}'"
        else:
            return f"UNCHANGED: '{key_str}: {self.build_value}'"


class ResourceYAMLDifference(YAMLWithComments[tuple[Union[str, int], ...], ResourceProperty]):
    """This represents a YAML file that contains resources and their properties.

    It is used to compare a local resource file with a CDF resource.
    """

    def __init__(
        self,
        items: dict[tuple[str | int, ...], ResourceProperty],
        comments: dict[tuple[str, ...], YAMLComment] | None = None,
    ) -> None:
        super().__init__(items or {})
        self._comments = comments or {}

    def _get_comment(self, key: tuple[str, ...]) -> YAMLComment | None:
        return self._comments.get(key)

    @classmethod
    def load(cls, build_content: str, source_content: str) -> Self:
        comments = cls._extract_comments(build_content)
        build = read_yaml_content(build_content)
        build_flatten = cls._flatten(build)
        items: dict[tuple[str | int, ...], ResourceProperty] = {}
        for key, value in build_flatten.items():
            items[key] = ResourceProperty(
                key_path=key,
                build_value=value,
            )

        source_content, variable_by_placeholder = cls._replace_variables(source_content)
        source = read_yaml_content(source_content)
        source_items = cls._flatten(source)
        for key, value in source_items.items():
            for placeholder, variable in variable_by_placeholder.items():
                if placeholder in str(value):
                    items[key].variables.append(
                        Variable(
                            placeholder=placeholder,
                            name=variable_by_placeholder[placeholder],
                            source_value=str(value),
                        )
                    )
        return cls(items, comments)

    @classmethod
    def _flatten(
        cls, raw: dict[str, Any] | list[dict[str, Any]]
    ) -> dict[tuple[str | int, ...], str | int | float | bool | None]:
        if isinstance(raw, dict):
            return cls._flatten_dict(raw)
        elif isinstance(raw, list):
            raise NotImplementedError()
        else:
            raise ValueError(f"Expected a dictionary or list, got {type(raw)}")

    @classmethod
    def _flatten_dict(
        cls, raw: dict[str, Any], key_path: tuple[str | int, ...] = ()
    ) -> dict[tuple[str | int, ...], str | int | float | bool | None]:
        items: dict[tuple[str | int, ...], str | int | float | bool | None] = {}
        for key, value in raw.items():
            if key == "scopes":
                # Hack to handle that scopes is a list variable
                items[(*key_path, key)] = value
            elif isinstance(value, dict):
                items.update(cls._flatten_dict(value, (*key_path, key)))
            elif isinstance(value, list):
                for i, item in enumerate(value):
                    if isinstance(item, dict):
                        items.update(cls._flatten_dict(item, (*key_path, key, i)))
                    else:
                        items[(*key_path, key, i)] = item
            else:
                items[(*key_path, key)] = value
        return items

    @classmethod
    def _replace_variables(cls, content: str) -> tuple[str, dict[str, str]]:
        variable_by_placeholder: dict[str, str] = {}
        seen: set[str] = set()
        for match in _VARIABLE_PATTERN.finditer(content):
            variable = match.group(1)
            if variable in seen:
                continue
            placeholder = f"VARIABLE_{uuid.uuid4().hex[:8]}"
            content = content.replace(f"{{{{{variable}}}}}", placeholder)
            variable_by_placeholder[placeholder] = variable
            seen.add(variable)
        return content, variable_by_placeholder

    def update_cdf_resource(self, cdf_resource: dict[str, Any]) -> None:
        for key, value in self._flatten_dict(cdf_resource).items():
            if key in self:
                self[key].cdf_value = value
            else:
                self[key] = ResourceProperty(key_path=key, cdf_value=value)

    def dump(self) -> dict[Any, Any]:
        dumped: dict[Any, Any] = {}
        for key, prop in self.items():
            current = dumped
            for part, next_part in itertools.pairwise(key):
                if isinstance(part, int) and isinstance(current, list) and len(current) < part + 1:
                    current.append({})
                    current = current[part]
                elif isinstance(part, int) and isinstance(current, list) and part < len(current):
                    current = current[part]
                elif isinstance(part, str) and isinstance(next_part, str):
                    current = current.setdefault(part, {})
                elif isinstance(part, str) and isinstance(next_part, int):
                    current = current.setdefault(part, [])
                else:
                    raise ValueError(f"Expected a string or int, got {type(part)}")
            if isinstance(key[-1], int) and isinstance(current, list):
                current.append(prop.value)
            elif isinstance(key[-1], str) and isinstance(current, dict):
                current[key[-1]] = prop.value
            else:
                raise ValueError(f"Expected a string or int, got {type(key[-1])}")
        return dumped

    def dump_yaml_with_comments(self, indent_size: int = 2) -> str:
        """Dump a config dictionary to a yaml string"""
        dumped_with_comments = self._dump_yaml_with_comments(indent_size, False)
        for key, prop in self.items():
            for variable in prop.variables:
                if variable.placeholder:
                    dumped_with_comments = dumped_with_comments.replace(
                        variable.placeholder, f"{{{{{variable.name}}}}}"
                    )
        return dumped_with_comments

    def display(self, title: str | None = None) -> None:
        added = [prop for prop in self.values() if prop.is_added]
        changed = [prop for prop in self.values() if prop.is_changed]
        cannot_change = [prop for prop in self.values() if prop.is_cannot_change]
        unchanged = [
            prop for prop in self.values() if not prop.is_added and not prop.is_changed and not prop.is_cannot_change
        ]

        content: list[str] = []
        if added:
            content.append("\n**Added properties**(Either set in CDF UI or default values set by CDF):")
            content.extend([f" - {prop}" for prop in added])
        if changed:
            content.append("\n**Changed properties:**")
            content.extend([f" - {prop}" for prop in changed])
        if cannot_change:
            content.append("\n**Cannot change properties**")
            content.extend([f" - {prop}" for prop in cannot_change])
        if unchanged:
            content.append(f"\n**{len(unchanged)} properties unchanged**")

        print(Panel.fit(Markdown("\n".join(content), justify="left"), title=title or "Resource differences"))


@dataclass
class Line:
    line_no: int
    build_value: str | None = None
    source_value: str | None = None
    cdf_value: str | None = None
    variables: list[str] | None = None

    @property
    def value(self) -> str:
        if self.variables:
            if self.source_value is None:
                raise ValueError("Source value should be set if there are variables")
            return self.source_value
        value = self.cdf_value or self.build_value
        if value is None:
            raise ValueError("CDF value or build value should be set")
        return value

    @property
    def is_changed(self) -> bool:
        return (
            self.build_value != self.cdf_value
            and self.build_value is not None
            and self.cdf_value is not None
            and self.variables is None
        )

    @property
    def is_added(self) -> bool:
        return self.build_value is None and self.cdf_value is not None

    @property
    def is_cannot_change(self) -> bool:
        return (
            self.build_value != self.cdf_value
            and self.variables is not None
            and self.build_value is not None
            and self.cdf_value is not None
        )


class TextFileDifference(UserList):
    def __init__(self, lines: list[Line] | None) -> None:
        super().__init__(lines or [])

    @classmethod
    def load(cls, build_content: str, source_content: str) -> Self:
        lines = []
        # Build and source content should have the same number of lines
        for no, (build, source) in enumerate(zip(build_content.splitlines(), source_content.splitlines())):
            variables = [v.group(1) for v in _VARIABLE_PATTERN.finditer(source)] or None
            lines.append(
                Line(
                    line_no=no + 1,
                    build_value=build,
                    source_value=source,
                    variables=variables,
                )
            )
        return cls(lines)

    def update_cdf_content(self, cdf_content: str) -> None:
        for i, line in enumerate(cdf_content.splitlines()):
            if i < len(self):
                self[i].cdf_value = line
            else:
                self.append(Line(cdf_value=line, line_no=i + 1))

    def dump(self) -> str:
        return "\n".join(line.value for line in self) + "\n"

    def display(self, title: str | None = None) -> None:
        added = [line for line in self if line.is_added]
        changed = [line for line in self if line.is_changed]
        cannot_change = [line for line in self if line.is_cannot_change]
        unchanged_count = len(self) - len(added) - len(changed) - len(cannot_change)

        content: list[str] = []
        if added:
            content.append("\n**Added lines**")
            if len(added) == 1:
                content.append(f" - Line {added[0].line_no}: '{added[0].cdf_value}'")
            else:
                content.append(f" - Line {added[0].line_no} - {added[-1].line_no}: {len(added)} lines")
        if changed:
            content.append("\n**Changed lines**")
            if len(changed) == 1:
                content.append(f" - Line {changed[0].line_no}: '{changed[0].source_value}' -> '{changed[0].cdf_value}'")
            else:
                content.append(f" - Line {changed[0].line_no} - {changed[-1].line_no}: {len(changed)} lines")
        if cannot_change:
            content.append("\n**Cannot change lines**")
            if len(cannot_change) == 1:
                content.append(
                    f" - Line {cannot_change[0].line_no}: '{cannot_change[0].source_value}' -> '{cannot_change[0].cdf_value}'"
                )
            else:
                content.append(
                    f" - Line {cannot_change[0].line_no} - {cannot_change[-1].line_no}: {len(cannot_change)} lines"
                )
        if unchanged_count != 0:
            content.append(f"\n**{unchanged_count} lines unchanged**")

        print(Panel.fit(Markdown("\n".join(content), justify="left"), title=title or "File differences"))


class PullCommand(ToolkitCommand):
    def __init__(self, print_warning: bool = True, skip_tracking: bool = False, silent: bool = False) -> None:
        super().__init__(print_warning, skip_tracking, silent)
        self._clean_command = CleanCommand(print_warning, skip_tracking=True)

    def pull_module(
        self,
        module_name_or_path: str | Path | None,
        organization_dir: Path,
        env: str,
        dry_run: bool,
        verbose: bool,
        env_vars: EnvironmentVariables,
    ) -> None:
        client = env_vars.get_client()
        client.config.is_strict_validation = False
        if not module_name_or_path:
            modules = ModuleDirectories.load(organization_dir, None)
            if not modules:
                raise ToolkitValueError(
                    "No module argument provided and no modules found in the organization directory."
                )

            selected = questionary.select(
                "Select a module to pull",
                choices=[Choice(title=module.name, value=module.name) for module in modules],
            ).ask()
        else:
            selected = parse_user_selected_modules([module_name_or_path])[0]
        build_module: str | Path
        if isinstance(selected, str):
            build_module = selected
        elif isinstance(selected, Path):
            try:
                # If the selected path is a sub-path of a module, we
                # need to build the entire module.
                build_module = module_directory_from_path(selected)
            except ValueError:
                # Remove this if-statement and set the build_module to selected
                # to support pulling more than one module.
                if is_module_path(selected):
                    build_module = selected
                else:
                    raise ToolkitValueError(
                        "Select module or a sub-path of a module. Multiple modules are not supported."
                    )
        else:
            raise ValueError("Expected a string or Path")
        build_cmd = BuildCommand(silent=True, skip_tracking=True)
        build_dir = Path(tempfile.mkdtemp())
        try:
            built_modules = build_cmd.execute(
                verbose=verbose,
                organization_dir=organization_dir,
                build_dir=build_dir,
                selected=[build_module],
                build_env_name=env,
                no_clean=False,
                client=client,
                on_error="raise",
            )
        except ToolkitError as e:
            raise ToolkitError(f"Failed to build module {module_name_or_path}.") from e
        else:
            self._pull_build_dir(build_dir, selected, built_modules, dry_run, env, client, env_vars)
        finally:
            try:
                safe_rmtree(build_dir)
            except Exception as e:
                raise ToolkitError(f"Failed to clean up temporary build directory {build_dir}.") from e

    def _pull_build_dir(
        self,
        build_dir: Path,
        selected: Path | str,
        built_modules: BuiltModuleList,
        dry_run: bool,
        build_env_name: str,
        client: ToolkitClient,
        env_vars: EnvironmentVariables,
    ) -> None:
        build_environment_file_path = build_dir / BUILD_ENVIRONMENT_FILE
        built = BuildEnvironment.load(read_yaml_file(build_environment_file_path), build_env_name, "pull")
        selected_loaders = self._clean_command.get_selected_loaders(
            build_dir, read_resource_folders=built.read_resource_folders, include=None
        )

        if len(selected_loaders) == 0:
            if isinstance(selected, Path):
                self.warn(LowSeverityWarning(f"No valid resource recognized at {selected.as_posix()}"))
            else:
                self.warn(LowSeverityWarning(f"No valid resources recognized in {selected}"))
            return

        results = DeployResults([], action="pull", dry_run=dry_run)
        for loader_cls in selected_loaders:
            if not issubclass(loader_cls, ResourceCRUD):
                continue
            loader = loader_cls.create_loader(client, build_dir)
            resources: BuiltFullResourceList[T_ID] = built_modules.get_resources(  # type: ignore[valid-type]
                None,
                loader.folder_name,  # type: ignore[arg-type]
                loader.kind,
                selected,
                is_supported_file=loader.is_supported_file,
            )
            if not resources:
                continue
            if isinstance(loader, HostedExtractorSourceCRUD | HostedExtractorDestinationCRUD):
                self.warn(
                    LowSeverityWarning(f"Skipping {loader.display_name} as it is not supported by the pull command.")
                )
                continue
            if isinstance(loader, GraphQLCRUD | FunctionCRUD | StreamlitCRUD):
                self.warn(
                    LowSeverityWarning(
                        f"Skipping {loader.display_name} as it is not supported by the pull command due to"
                        "the external file(s)."
                    )
                )
                continue
            if isinstance(loader, GroupAllScopedCRUD):
                # We have two loaders for Groups. We skip this one and
                # only use the GroupResourceScopedLoader
                continue

            result = self._pull_resources(loader, resources, dry_run, env_vars.dump(include_os=True))
            results[loader.display_name] = result

        table = results.counts_table(exclude_columns={"Total"})
        print(table)

    def _pull_resources(
        self,
        loader: ResourceCRUD[
            T_ID, T_ResourceRequest, T_ResourceResponse, T_ResourceRequestList, T_ResourceResponseList
        ],
        resources: BuiltFullResourceList[T_ID],
        dry_run: bool,
        environment_variables: dict[str, str | None],
    ) -> ResourceDeployResult:
        cdf_resources = loader.retrieve(resources.identifiers)
        cdf_resource_by_id: dict[T_ID, T_ResourceResponse] = {loader.get_id(r): r for r in cdf_resources}

        resources_by_file = resources.by_file()
        file_results = ResourceDeployResult(loader.display_name)
        environment_variables = environment_variables or {}
        for source_file, resources in resources_by_file.items():
            local_resource_by_id = self._get_local_resource_dict_by_id(resources, loader, environment_variables)
            has_changes, to_write = self._get_to_write(local_resource_by_id, cdf_resource_by_id, file_results, loader)

            if has_changes and not dry_run:
                new_content, extra_files = self._to_write_content(
                    safe_read(source_file), to_write, resources, environment_variables, loader, source_file
                )
                with source_file.open("w", encoding=ENCODING, newline=NEWLINE) as f:
                    f.write(new_content)
                for filepath, content in extra_files.items():
                    with filepath.open("w", encoding=ENCODING, newline=NEWLINE) as f:
                        f.write(content)

        return file_results

    def _get_to_write(
        self,
        local_resource_by_id: dict[T_ID, dict[str, Any]],
        cdf_resource_by_id: dict[T_ID, T_ResourceResponse],
        file_results: ResourceDeployResult,
        loader: ResourceCRUD[
            T_ID, T_ResourceRequest, T_ResourceResponse, T_ResourceRequestList, T_ResourceResponseList
        ],
    ) -> tuple[bool, dict[T_ID, dict[str, Any]]]:
        to_write: dict[T_ID, dict[str, Any]] = {}
        has_changes = False
        for item_id, local_dict in local_resource_by_id.items():
            cdf_resource = cdf_resource_by_id.get(item_id)
            if cdf_resource is None:
                file_results.unchanged += 1
                to_write[item_id] = local_dict
                self.warn(
                    MediumSeverityWarning(
                        f"No {loader.display_name} with id {item_id} found in CDF. Have you deployed it?"
                    )
                )
                continue
            cdf_dumped = loader.dump_resource(cdf_resource, local_dict)

            if cdf_dumped == local_dict:
                file_results.unchanged += 1
                to_write[item_id] = local_dict
            else:
                file_results.changed += 1
                to_write[item_id] = cdf_dumped
                has_changes = True
        return has_changes, to_write

    @staticmethod
    def _get_local_resource_dict_by_id(
        resources: BuiltFullResourceList[T_ID],
        loader: ResourceCRUD[
            T_ID, T_ResourceRequest, T_ResourceResponse, T_ResourceRequestList, T_ResourceResponseList
        ],
        environment_variables: dict[str, str | None],
    ) -> dict[T_ID, dict[str, Any]]:
        unique_destinations = {r.destination for r in resources if r.destination}
        local_resource_by_id: dict[T_ID, dict[str, Any]] = {}
        local_resource_ids = set(resources.identifiers)
        for destination in unique_destinations:
            resource_list = loader.load_resource_file(destination, environment_variables)
            for resource_dict in resource_list:
                identifier = loader.get_id(resource_dict)
                if identifier in local_resource_ids:
                    local_resource_by_id[identifier] = resource_dict
        return local_resource_by_id

    @staticmethod
    def _select_resource_ids(
        all_: bool, id_: T_ID, loader: ResourceCRUD, local_resources: BuiltFullResourceList, organization_dir: Path
    ) -> BuiltFullResourceList[T_ID]:
        if all_:
            return local_resources
        if id_ is None:
            return questionary.select(
                f"Select a {loader.display_name} to pull",
                choices=[Choice(title=f"{r.identifier!r} - ({r.module_name})", value=r) for r in local_resources],
            ).ask()
        if id_ not in local_resources.identifiers:
            raise ToolkitMissingResourceError(
                f"No {loader.display_name} with external id {id_} found in the current configuration in {organization_dir}."
            )
        return BuiltFullResourceList([r for r in local_resources if r.identifier == id_])

    def _to_write_content(
        self,
        source: str,
        to_write: dict[T_ID, dict[str, Any]],
        resources: BuiltFullResourceList[T_ID],
        environment_variables: dict[str, str | None],
        loader: ResourceCRUD[
            T_ID, T_ResourceRequest, T_ResourceResponse, T_ResourceRequestList, T_ResourceResponseList
        ],
        source_file: Path,
    ) -> tuple[str, dict[Path, str]]:
        # 1. Replace all variables with placeholders
        # 2. Load source and keep the comments
        # 3. Update the to_write dict with the placeholders
        # 4. Dump the yaml with the placeholders
        # 5. Replace the placeholders with the variables
        # 6. Add the comments back

        # All resources are assumed to be in the same file, and thus the same build variables.
        variables = resources[0].build_variables
        if environment_variables:
            variables_with_environment_list: list[BuildVariable] = []
            for variable in variables:
                if isinstance(variable.value, str) and ENV_VAR_PATTERN.match(variable.value):
                    for key, value in environment_variables.items():
                        if key in variable.value and isinstance(value, str):
                            # Running through all environment variables, in case multiple are used in the same variable.
                            # Note that variable are immutable, so we are not modifying the original variable.
                            variable = dataclasses.replace(variable, value=variable.value.replace(f"${{{key}}}", value))
                    variables_with_environment_list.append(variable)
                elif isinstance(variable.value, tuple):
                    new_value: list[str | int | float | bool] = []
                    for var_item in variable.value:
                        if isinstance(var_item, str) and ENV_VAR_PATTERN.match(var_item):
                            for key, value in environment_variables.items():
                                if key in var_item and isinstance(value, str):
                                    var_item = var_item.replace(f"${{{key}}}", value)
                        new_value.append(var_item)
                    variables_with_environment_list.append(dataclasses.replace(variable, value=tuple(new_value)))  # type: ignore[arg-type]
                else:
                    variables_with_environment_list.append(variable)
            variables = BuildVariables(variables_with_environment_list)

        content, value_by_placeholder = variables.replace(source, source_file, use_placeholder=True)
        comments = YAMLComments.load(source)
        # If there is a variable in the identifier, we need to replace it with the value
        # such that we can look it up in the to_write dict.
        if isinstance(loader, ExtractionPipelineConfigCRUD):
            # The safe read in ExtractionPipelineConfigLoader stringifies the config dict,
            # but we need to load it as a dict so we can write it back to the file maintaining
            # the order or the keys.
            loaded = read_yaml_content(variables.replace(source, source_file))
            loaded_with_placeholder = read_yaml_content(content)
        else:
            loaded = read_yaml_content(loader.safe_read(variables.replace(source, source_file)))
            loaded_with_placeholder = read_yaml_content(loader.safe_read(content))

        built_by_identifier = {r.identifier: r for r in resources}
        updated: dict[str, Any] | list[dict[str, Any]]
        extra_files: dict[Path, str] = {}
        replacer = ResourceReplacer(value_by_placeholder, loader)
        if isinstance(loaded, dict) and isinstance(loaded_with_placeholder, dict):
            item_id = loader.get_id(loaded)
            updated = self._update(
                item_id,
                loaded,
                loaded_with_placeholder,
                to_write,
                built_by_identifier,
                replacer,
                extra_files,
            )
        elif isinstance(loaded, list) and isinstance(loaded_with_placeholder, list):
            updated = []
            for i, item in enumerate(loaded):
                item_id = loader.get_id(item)
                updated.append(
                    self._update(
                        item_id,
                        item,
                        loaded_with_placeholder[i],
                        to_write,
                        built_by_identifier,
                        replacer,
                        extra_files,
                    )
                )
        else:
            raise ValueError("Loaded and loaded_with_ids should be of the same type")

        dumped = yaml_safe_dump(updated)
        for placeholder, variable in value_by_placeholder.items():
            dumped = dumped.replace(placeholder, f"{{{{ {variable.key} }}}}")
        file_content = comments.dump(dumped)
        return file_content, extra_files

    @classmethod
    def _update(
        cls,
        item_id: T_ID,
        loaded: dict[str, Any],
        loaded_with_placeholder: dict[str, Any],
        to_write: dict[T_ID, dict[str, Any]],
        built_by_identifier: dict[T_ID, BuiltResourceFull[T_ID]],
        replacer: "ResourceReplacer",
        extra_files: dict[Path, str],
    ) -> dict[str, Any]:
        if item_id not in to_write:
            raise ToolkitMissingResourceError(f"Resource {item_id} not found in to_write.")
        item_write = to_write[item_id]
        if item_id not in built_by_identifier:
            raise ToolkitMissingResourceError(f"Resource {item_id} not found in resources.")
        built = built_by_identifier[item_id]
        if built.extra_sources:
            builder = create_builder(built.resource_dir, None)
            for extra in built.extra_sources:
                extra_content, extra_placeholders = built.build_variables.replace(
                    safe_read(extra.path), extra.path, use_placeholder=True
                )
                key, _ = builder.load_extra_field(extra_content)
                if key in item_write:
                    new_extra = item_write.pop(key)
                    for placeholder, variable in extra_placeholders.items():
                        if placeholder in extra_content:
                            new_extra = new_extra.replace(variable.value, f"{{{{ {variable.key} }}}}")
                    extra_files[extra.path] = new_extra
        return replacer.replace(loaded, loaded_with_placeholder, item_write)


class ResourceReplacer:
    """Replaces values in a local resource directory with the updated values from CDF.

    The local resource dict order is maintained. In addition, placeholders are used for variables.
    """

    def __init__(self, value_by_placeholder: dict[str, BuildVariable], loader: ResourceCRUD) -> None:
        self._value_by_placeholder = value_by_placeholder
        self._loader = loader

    def replace(
        self,
        current: dict[str, Any],
        placeholder: dict[str, Any],
        to_write: dict[str, Any],
    ) -> dict[str, Any]:
        return self._replace_dict(current, placeholder, to_write, tuple())

    def _replace_dict(
        self,
        current: dict[str, Any],
        placeholder: dict[str, Any],
        to_write: dict[str, Any],
        json_path: tuple[str | int, ...],
    ) -> dict[str, Any]:
        # Modified first to maintain original order
        # Then added, and skip removed
        updated: dict[str, Any] = {}
        variable_keys: set[str] = set()
        for modified_key, current_value in current.items():
            if modified_key not in to_write:
                # Removed item by skipping
                continue
            cdf_value = to_write[modified_key]

            if modified_key in placeholder:
                placeholder_value = placeholder[modified_key]
            elif variable_key := next(
                (
                    key
                    for key, variable in self._value_by_placeholder.items()
                    if key in placeholder and variable.value == modified_key
                ),
                None,
            ):
                # The key is a variable
                variable_keys.add(modified_key)
                modified_key = variable_key
                placeholder_value = placeholder[variable_key]
            else:
                # Bug in the code if this is reached, using a fallback.
                placeholder_value = current_value

            if isinstance(current_value, dict) and isinstance(cdf_value, dict):
                updated[modified_key] = self._replace_dict(
                    current_value, placeholder_value, cdf_value, (*json_path, modified_key)
                )
            elif isinstance(current_value, list) and isinstance(cdf_value, list):
                if isinstance(placeholder_value, str) and current_value == cdf_value:
                    # A list variable is used, and the list is unchanged.
                    updated[modified_key] = placeholder_value
                elif isinstance(placeholder_value, list):
                    updated[modified_key] = self._replace_list(
                        current_value, placeholder_value, cdf_value, (*json_path, modified_key)
                    )
                else:
                    # A list variable is used, but the list is changed. Since the value is represented as a single
                    # string, we cannot update it.
                    if variable := self._value_by_placeholder.get(placeholder_value):
                        raise ToolkitValueError(
                            f"Pull is not supported for list variable: {variable.key}: {variable.value_variable}"
                        )
                    raise ToolkitValueError("Pull is not supported for list variable.")
            else:
                updated[modified_key] = self._replace_value(
                    current_value, placeholder_value, cdf_value, (*json_path, modified_key)
                )

        for new_key in to_write:
            if (new_key not in current) and new_key not in variable_keys:
                # Note there cannot be variables in new items
                updated[new_key] = to_write[new_key]
        return updated

    def _replace_list(
        self,
        current: list[Any],
        placeholder: list[Any],
        to_write: list[Any],
        json_path: tuple[str | int, ...],
    ) -> list[Any]:
        compare_indices, added_indices = self._loader.diff_list(current, to_write, json_path)
        updated: list[Any] = []
        for no, current_item in enumerate(current):
            if no not in compare_indices:
                # Removed item
                continue
            current_value = current_item
            placeholder_value = placeholder[no]
            cdf_value = to_write[compare_indices[no]]
            updated.append(self._replace_value(current_value, placeholder_value, cdf_value, (*json_path, no)))
        for added_index in added_indices:
            # Note there cannot be variables in new items
            updated.append(to_write[added_index])
        return updated

    def _replace_value(
        self,
        current: Any,
        placeholder_value: Any,
        to_write: Any,
        json_path: tuple[str | int, ...],
    ) -> Any:
        if isinstance(current, dict) and isinstance(to_write, dict):
            return self._replace_dict(current, placeholder_value, to_write, json_path)
        elif isinstance(current, list) and isinstance(to_write, list):
            return self._replace_list(current, placeholder_value, to_write, json_path)
        elif type(current) is type(to_write):
            if to_write == current:
                return placeholder_value
            if not isinstance(to_write, str):
                # Variable substitution is only supported for strings
                return to_write
            for placeholder, variable in self._value_by_placeholder.items():
                if placeholder in placeholder_value:
                    # We use the placeholder and not the {{ variable }} in the value to ensure
                    # that the result is valid yaml.
                    to_write = to_write.replace(variable.value, placeholder)  # type: ignore[arg-type]
                    # Iterate through all variables in case multiple are used in the same value.
            return to_write
        elif isinstance(current, dict) and isinstance(to_write, str):
            # This is a special case for the ExtractionPipelineConfigLoader where the config dict is typically a
            # dict locally, but returned as a string from the server.
            try:
                to_write = read_yaml_content(to_write)
            except yaml.YAMLError:
                ...
            else:
                return self._replace_dict(current, placeholder_value, to_write, json_path)

        raise ToolkitValueError(
            f"CDF value and local value should be of the same type in {'.'.join(map(str, json_path))}, "
            f"got {type(current)} != {type(to_write)}"
        )



from ._base import CellValue, Chunk, DataType, PrimaryCellValue, SchemaColumn
from ._compression import (
    COMPRESSION_BY_NAME,
    COMPRESSION_BY_SUFFIX,
    Compression,
    GzipCompression,
    Uncompressed,
)
from ._readers import (
    FILE_READ_CLS_BY_FORMAT,
    TABLE_READ_CLS_BY_FORMAT,
    CSVReader,
    FailedParsing,
    FileReader,
    NDJsonReader,
    ParquetReader,
    YAMLReader,
    YMLReader,
)
from ._writers import (
    FILE_WRITE_CLS_BY_FORMAT,
    TABLE_WRITE_CLS_BY_FORMAT,
    CSVWriter,
    FileWriter,
    NDJsonWriter,
    ParquetWriter,
    YAMLWriter,
    YMLWriter,
)

__all__ = [
    "COMPRESSION_BY_NAME",
    "COMPRESSION_BY_SUFFIX",
    "FILE_READ_CLS_BY_FORMAT",
    "FILE_WRITE_CLS_BY_FORMAT",
    "TABLE_READ_CLS_BY_FORMAT",
    "TABLE_WRITE_CLS_BY_FORMAT",
    "CSVReader",
    "CSVWriter",
    "CellValue",
    "Chunk",
    "Compression",
    "DataType",
    "FailedParsing",
    "FileReader",
    "FileWriter",
    "GzipCompression",
    "NDJsonReader",
    "NDJsonWriter",
    "ParquetReader",
    "ParquetWriter",
    "PrimaryCellValue",
    "SchemaColumn",
    "Uncompressed",
    "YAMLReader",
    "YAMLWriter",
    "YMLReader",
    "YMLWriter",
]



from typing import Literal

from ._base import DataSelector, SelectorObject


class SelectedTable(SelectorObject):
    """Selected RAW table"""

    db_name: str
    table_name: str


class RawTableSelector(DataSelector):
    type: Literal["rawTable"] = "rawTable"
    kind: Literal["RawRows"] = "RawRows"
    table: SelectedTable
    key: str | None = None

    @property
    def group(self) -> str:
        return self.table.db_name

    def __str__(self) -> str:
        return self.table.table_name



from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import TYPE_CHECKING, overload

from cognite.client import ClientConfig
from cognite.client.data_classes.capabilities import (
    AssetsAcl,
    Capability,
    DataSetsAcl,
    EventsAcl,
    ExtractionPipelinesAcl,
    FilesAcl,
    FunctionsAcl,
    LocationFiltersAcl,
    SecurityCategoriesAcl,
    TimeSeriesAcl,
)
from cognite.client.exceptions import CogniteAPIError
from cognite.client.utils.useful_types import SequenceNotStr
from rich.console import Console

from cognite_toolkit._cdf_tk.client.api_client import ToolkitAPI
from cognite_toolkit._cdf_tk.constants import DRY_RUN_ID
from cognite_toolkit._cdf_tk.tk_warnings import MediumSeverityWarning
from cognite_toolkit._cdf_tk.utils import humanize_collection

if TYPE_CHECKING:
    from cognite_toolkit._cdf_tk.client._toolkit_client import ToolkitClient


class LookUpAPI(ToolkitAPI, ABC):
    dry_run_id: int = DRY_RUN_ID

    def __init__(
        self, config: ClientConfig, api_version: str | None, cognite_client: "ToolkitClient", console: Console
    ) -> None:
        super().__init__(config, api_version, cognite_client)
        self._console = console
        self._cache: dict[str, int | None] = {}
        self._reverse_cache: dict[int, str | None] = {}

    @property
    def resource_name(self) -> str:
        return type(self).__name__.removesuffix("LookUpAPI")

    @overload
    def id(self, external_id: str, is_dry_run: bool = False, allow_empty: bool = False) -> int | None: ...

    @overload
    def id(
        self, external_id: SequenceNotStr[str], is_dry_run: bool = False, allow_empty: bool = False
    ) -> list[int]: ...

    def id(
        self, external_id: str | SequenceNotStr[str], is_dry_run: bool = False, allow_empty: bool = False
    ) -> int | None | list[int]:
        """Lookup internal IDs for given external IDs.

        Args:
            external_id: A string external ID or a sequence of string external IDs to look up.
            is_dry_run: If True, return a predefined dry run ID for missing external IDs instead of raising an error.
            allow_empty: If True, allow empty string external IDs and return 0 for them.

        Returns:
            The corresponding internal ID as an integer if a single external ID is provided,
            or a list of internal IDs if a sequence of external IDs is provided.
        """
        ids = [external_id] if isinstance(external_id, str) else external_id
        need_lookup = [id for id in ids if id not in self._cache]
        if allow_empty and "" in need_lookup:
            # Note we do not want to put empty string in the cache. It is a special case that
            # as of 01/02/2025 only applies to LocationFilters
            need_lookup.remove("")
        if need_lookup:
            self._do_lookup_external_ids(need_lookup, is_dry_run)

        if isinstance(external_id, str):
            return self._get_id_from_cache(external_id, is_dry_run, allow_empty)
        else:
            internal_ids = (
                self._get_id_from_cache(external_id, is_dry_run, allow_empty) for external_id in external_id
            )
            return [id_ for id_ in internal_ids if id_ is not None]

    def _do_lookup_external_ids(self, external_ids: list[str], is_dry_run: bool) -> None:
        try:
            ids_by_external_id = self._id(external_ids)
        except CogniteAPIError as e:
            if 400 <= e.code < 500:
                missing_capabilities = self._toolkit_client.verify.authorization(self._read_acl())
                if missing_capabilities:
                    raise self._toolkit_client.verify.create_error(
                        missing_capabilities,
                        f"lookup {self.resource_name} with external_id {external_ids}",
                    )
            # Raise the original error if it's not a 400 or the user has access to read the resource.from
            raise
        self._cache.update(ids_by_external_id)
        self._reverse_cache.update({v: k for k, v in ids_by_external_id.items()})
        missing_external_ids = [ext_id for ext_id in external_ids if ext_id not in ids_by_external_id]
        if missing_external_ids and not is_dry_run:
            plural = "s" if len(missing_external_ids) > 1 else ""
            plural2 = "They do" if len(missing_external_ids) > 1 else "It does"
            MediumSeverityWarning(
                f"Failed to retrieve {self.resource_name} with external_id{plural} "
                f"{humanize_collection(missing_external_ids)}. {plural2} not exist in CDF"
            ).print_warning(console=self._console)
            self._cache.update({ext_id: None for ext_id in missing_external_ids})

    def _get_id_from_cache(self, external_id: str, is_dry_run: bool = False, allow_empty: bool = False) -> int | None:
        if allow_empty and external_id == "":
            return 0
        elif is_dry_run:
            return self._cache.get(external_id, self.dry_run_id)
        else:
            return self._cache[external_id]

    @overload
    def external_id(self, id: int) -> str | None: ...

    @overload
    def external_id(self, id: Sequence[int]) -> list[str]: ...

    def external_id(self, id: int | Sequence[int]) -> str | None | list[str]:
        """Lookup external IDs for given internal IDs.

        Args:
            id: An integer ID or a sequence of integer IDs to look up. Note that an ID of 0 corresponds
                to an empty string external ID.

        Returns:
            The corresponding external ID as a string if a single ID is provided,
            or a list of external IDs if a sequence of IDs is provided.
            If an ID does not exist, None is returned for that ID.

        """
        ids = [id] if isinstance(id, int) else id
        need_lookup = [id_ for id_ in ids if id_ not in self._reverse_cache if id_ != 0]
        if need_lookup:
            self._do_lookup_internal_ids(need_lookup)

        if isinstance(id, int):
            return self._get_external_id_from_cache(id)
        else:
            external_ids = (self._get_external_id_from_cache(id_) for id_ in ids)
            return [id_ for id_ in external_ids if id_ is not None]

    def _do_lookup_internal_ids(self, ids: list[int]) -> None:
        try:
            found_by_id = self._external_id(ids)
        except CogniteAPIError as e:
            if 400 <= e.code < 500:
                missing_capabilities = self._toolkit_client.verify.authorization(self._read_acl())
                if missing_capabilities:
                    raise self._toolkit_client.verify.create_error(
                        missing_capabilities,
                        f"lookup {self.resource_name} with id {ids}",
                    )
            # Raise the original error if it's not a 400 or the user has access to read the resource.from
            raise
        self._reverse_cache.update(found_by_id)
        self._cache.update({v: k for k, v in found_by_id.items()})
        missing_ids = [id for id in ids if id not in found_by_id]
        if not missing_ids:
            return None
        plural = "s" if len(missing_ids) > 1 else ""
        plural2 = "They do" if len(missing_ids) > 1 else "It does"
        MediumSeverityWarning(
            f"Failed to retrieve {self.resource_name} with id{plural} "
            f"{humanize_collection(missing_ids)}. {plural2} not exist in CDF"
        ).print_warning(console=self._console)
        # Cache the missing IDs with None to avoid repeated lookups
        self._reverse_cache.update({missing_id: None for missing_id in missing_ids})
        return None

    def _get_external_id_from_cache(self, id: int) -> str | None:
        if id == 0:
            # Reverse of looking up an empty string.
            return ""
        return self._reverse_cache.get(id)

    @abstractmethod
    def _id(self, external_id: SequenceNotStr[str]) -> dict[str, int]:
        raise NotImplementedError

    @abstractmethod
    def _external_id(self, id: Sequence[int]) -> dict[int, str]:
        raise NotImplementedError

    @abstractmethod
    def _read_acl(self) -> Capability:
        raise NotImplementedError


class DataSetLookUpAPI(LookUpAPI):
    def _id(self, external_id: SequenceNotStr[str]) -> dict[str, int]:
        return {
            data_set.external_id: data_set.id
            for data_set in self._cognite_client.data_sets.retrieve_multiple(
                external_ids=external_id, ignore_unknown_ids=True
            )
            if data_set.external_id and data_set.id
        }

    def _external_id(self, id: Sequence[int]) -> dict[int, str]:
        return {
            data_set.id: data_set.external_id
            for data_set in self._cognite_client.data_sets.retrieve_multiple(ids=id, ignore_unknown_ids=True)
            if data_set.external_id and data_set.id
        }

    def _read_acl(self) -> Capability:
        return DataSetsAcl(
            [DataSetsAcl.Action.Read],
            scope=DataSetsAcl.Scope.All(),
        )


class AssetLookUpAPI(LookUpAPI):
    def _id(self, external_id: SequenceNotStr[str]) -> dict[str, int]:
        return {
            asset.external_id: asset.id
            for asset in self._cognite_client.assets.retrieve_multiple(
                external_ids=external_id, ignore_unknown_ids=True
            )
            if asset.external_id and asset.id
        }

    def _external_id(self, id: Sequence[int]) -> dict[int, str]:
        return {
            asset.id: asset.external_id
            for asset in self._cognite_client.assets.retrieve_multiple(ids=id, ignore_unknown_ids=True)
            if asset.external_id and asset.id
        }

    def _read_acl(self) -> Capability:
        return AssetsAcl(
            [AssetsAcl.Action.Read],
            scope=AssetsAcl.Scope.All(),
        )


class TimeSeriesLookUpAPI(LookUpAPI):
    def _id(self, external_id: SequenceNotStr[str]) -> dict[str, int]:
        return {
            ts.external_id: ts.id
            for ts in self._cognite_client.time_series.retrieve_multiple(
                external_ids=external_id, ignore_unknown_ids=True
            )
            if ts.external_id and ts.id
        }

    def _external_id(self, id: Sequence[int]) -> dict[int, str]:
        return {
            ts.id: ts.external_id
            for ts in self._cognite_client.time_series.retrieve_multiple(ids=id, ignore_unknown_ids=True)
            if ts.external_id and ts.id
        }

    def _read_acl(self) -> Capability:
        return TimeSeriesAcl(
            [TimeSeriesAcl.Action.Read],
            scope=TimeSeriesAcl.Scope.All(),
        )


class FileMetadataLookUpAPI(LookUpAPI):
    def _id(self, external_id: SequenceNotStr[str]) -> dict[str, int]:
        return {
            file.external_id: file.id
            for file in self._cognite_client.files.retrieve_multiple(external_ids=external_id, ignore_unknown_ids=True)
            if file.external_id and file.id
        }

    def _external_id(self, id: Sequence[int]) -> dict[int, str]:
        return {
            file.id: file.external_id
            for file in self._cognite_client.files.retrieve_multiple(ids=id, ignore_unknown_ids=True)
            if file.external_id and file.id
        }

    def _read_acl(self) -> Capability:
        return FilesAcl(
            [FilesAcl.Action.Read],
            scope=FilesAcl.Scope.All(),
        )


class EventLookUpAPI(LookUpAPI):
    def _id(self, external_id: SequenceNotStr[str]) -> dict[str, int]:
        return {
            event.external_id: event.id
            for event in self._cognite_client.events.retrieve_multiple(
                external_ids=external_id, ignore_unknown_ids=True
            )
            if event.external_id and event.id
        }

    def _external_id(self, id: Sequence[int]) -> dict[int, str]:
        return {
            event.id: event.external_id
            for event in self._cognite_client.events.retrieve_multiple(ids=id, ignore_unknown_ids=True)
            if event.external_id and event.id
        }

    def _read_acl(self) -> Capability:
        return EventsAcl(
            [EventsAcl.Action.Read],
            scope=EventsAcl.Scope.All(),
        )


class ExtractionPipelineLookUpAPI(LookUpAPI):
    def _id(self, external_id: SequenceNotStr[str]) -> dict[str, int]:
        return {
            pipeline.external_id: pipeline.id
            for pipeline in self._cognite_client.extraction_pipelines.retrieve_multiple(
                external_ids=external_id, ignore_unknown_ids=True
            )
            if pipeline.external_id and pipeline.id
        }

    def _external_id(self, id: Sequence[int]) -> dict[int, str]:
        return {
            pipeline.id: pipeline.external_id
            for pipeline in self._cognite_client.extraction_pipelines.retrieve_multiple(ids=id, ignore_unknown_ids=True)
            if pipeline.external_id and pipeline.id
        }

    def _read_acl(self) -> Capability:
        return ExtractionPipelinesAcl(
            [ExtractionPipelinesAcl.Action.Read],
            scope=ExtractionPipelinesAcl.Scope.All(),
        )


class FunctionLookUpAPI(LookUpAPI):
    def _id(self, external_id: SequenceNotStr[str]) -> dict[str, int]:
        return {
            function.external_id: function.id
            for function in self._cognite_client.functions.retrieve_multiple(
                external_ids=external_id, ignore_unknown_ids=True
            )
            if function.external_id and function.id
        }

    def _external_id(self, id: Sequence[int]) -> dict[int, str]:
        return {
            function.id: function.external_id
            for function in self._cognite_client.functions.retrieve_multiple(ids=id, ignore_unknown_ids=True)
            if function.external_id and function.id
        }

    def _read_acl(self) -> Capability:
        return FunctionsAcl(
            [FunctionsAcl.Action.Read],
            scope=FunctionsAcl.Scope.All(),
        )


class AllLookUpAPI(LookUpAPI, ABC):
    def __init__(
        self, config: ClientConfig, api_version: str | None, cognite_client: "ToolkitClient", console: Console
    ) -> None:
        super().__init__(config, api_version, cognite_client, console)
        self._has_looked_up = False

    @abstractmethod
    def _lookup(self) -> None:
        raise NotImplementedError

    def _id(self, external_id: SequenceNotStr[str]) -> dict[str, int]:
        if not self._has_looked_up:
            self._lookup()
        found_pairs = ((ext_id, self._cache[ext_id]) for ext_id in external_id if ext_id in self._cache)
        return {k: v for k, v in found_pairs if v is not None}

    def _external_id(self, id: Sequence[int]) -> dict[int, str]:
        if not self._has_looked_up:
            self._lookup()
            self._has_looked_up = True
        found_pairs = ((id_, self._reverse_cache[id_]) for id_ in id if id_ in self._reverse_cache)
        return {k: v for k, v in found_pairs if v is not None}


class SecurityCategoriesLookUpAPI(AllLookUpAPI):
    def _lookup(self) -> None:
        categories = self._cognite_client.iam.security_categories.list(limit=-1)
        self._cache = {category.name: category.id for category in categories if category.name and category.id}
        self._reverse_cache = {category.id: category.name for category in categories if category.name and category.id}

    def name(self, id: int | Sequence[int]) -> str | list[str] | None:
        return self.external_id(id)

    def _read_acl(self) -> Capability:
        return SecurityCategoriesAcl(
            [SecurityCategoriesAcl.Action.List],
            scope=SecurityCategoriesAcl.Scope.All(),
        )


class LocationFiltersLookUpAPI(AllLookUpAPI):
    def _lookup(self) -> None:
        for location in self._toolkit_client.search.locations.list():
            if location.external_id and location.id:
                self._cache[location.external_id] = location.id
                self._reverse_cache[location.id] = location.external_id

    def _read_acl(self) -> Capability:
        return LocationFiltersAcl(
            [LocationFiltersAcl.Action.Read],
            scope=LocationFiltersAcl.Scope.All(),
        )

    def _id(self, external_id: SequenceNotStr[str]) -> dict[str, int]:
        if not self._has_looked_up:
            self._lookup()
        found_pairs = ((ext_id, self._cache[ext_id]) for ext_id in external_id if ext_id in self._cache)
        return {k: v for k, v in found_pairs if v is not None}


class LookUpGroup(ToolkitAPI):
    def __init__(
        self, config: ClientConfig, api_version: str | None, cognite_client: "ToolkitClient", console: Console
    ) -> None:
        super().__init__(config, api_version, cognite_client)
        self.data_sets = DataSetLookUpAPI(config, api_version, cognite_client, console)
        self.assets = AssetLookUpAPI(config, api_version, cognite_client, console)
        self.time_series = TimeSeriesLookUpAPI(config, api_version, cognite_client, console)
        self.files = FileMetadataLookUpAPI(config, api_version, cognite_client, console)
        self.events = EventLookUpAPI(config, api_version, cognite_client, console)
        self.security_categories = SecurityCategoriesLookUpAPI(config, api_version, cognite_client, console)
        self.location_filters = LocationFiltersLookUpAPI(config, api_version, cognite_client, console)
        self.extraction_pipelines = ExtractionPipelineLookUpAPI(config, api_version, cognite_client, console)
        self.functions = FunctionLookUpAPI(config, api_version, cognite_client, console)



from collections.abc import Iterable

import pytest
import yaml

from cognite_toolkit._cdf_tk.data_classes import YAMLComments
from cognite_toolkit._cdf_tk.utils.file import yaml_safe_dump


def load_dump_test_cases() -> Iterable:
    yield pytest.param("""key: value # comment""", 2, id="simple")
    yield pytest.param(
        """- item: 23 # comment
  # other comment
- item: 24""",
        2,
        id="list",
    )
    yield pytest.param(
        """first: value
second:
    key: value
    # above comment
    third:
    -   item: 23
    -   item: 24 # with comment""",
        4,
        id="nested with indent 4",
    )
    yield pytest.param(
        """- name: daily-8am-utc
  functionExternalId: fn_first_function
  description: Run every day at 8am UTC
  cronExpression: 0 8 * * *
  data:
    breakfast: 'today: peanut butter sandwich and coffee'
    lunch: 'today: greek salad and water'
    dinner: 'today: steak and red wine'
  authentication:
    # Credentials to use to run the function in this schedule.
    # In this example, we just use the main deploy credentials, so the result is the same, but use a different set of
    # credentials (env variables) if you want to run the function with different permissions.
    clientId: some_id
    clientSecret: some_secret""",
        2,
        id="realistic",
    )


class TestYAMLComments:
    @pytest.mark.parametrize("yaml_str, indent", list(load_dump_test_cases()))
    def test_load_dump(self, yaml_str: str, indent: int) -> None:
        comments = YAMLComments.load(yaml_str)
        data = yaml.safe_load(yaml_str)
        dumped = yaml_safe_dump(data, indent=indent)
        assert comments.dump(dumped) == yaml_str



from typing import TypeVar

from cognite.client.data_classes.capabilities import (
    AllScope,
    AppConfigScope,
    AssetRootIDScope,
    Capability,
    CurrentUserScope,
    DataSetScope,
    ExperimentsScope,
    ExtractionPipelineScope,
    IDScope,
    IDScopeLowerCase,
    InstancesScope,
    LegacyDataModelScope,
    LegacySpaceScope,
    PartitionScope,
    PostgresGatewayUsersScope,
    SpaceIDScope,
    TableScope,
)

T_Scope = TypeVar("T_Scope", bound=Capability.Scope)


def scope_intersection(scope1: T_Scope, scope2: T_Scope | None) -> Capability.Scope | None:
    """Returns the intersection of two scopes.
    If both scopes are None, returns None.
    If one scope is None, returns None (intersection with nothing is nothing).
    If both scopes are the same, returns that scope.
    If scopes are different types, raise an ValueError.
    """
    if scope2 is None:
        return None
    if type(scope1) is not type(scope2):
        raise ValueError("Cannot intersect scopes of different types")
    if isinstance(scope1, AllScope) and isinstance(scope2, AllScope):
        return AllScope()
    elif isinstance(scope1, AppConfigScope) and isinstance(scope2, AppConfigScope):
        if apps := sorted(set(scope1.apps) & set(scope2.apps)):
            return AppConfigScope(apps=apps)
        return None
    elif isinstance(scope1, DataSetScope) and isinstance(scope2, DataSetScope):
        if ids := sorted(set(scope1.ids) & set(scope2.ids)):
            return DataSetScope(ids=ids)
        return None
    elif isinstance(scope1, IDScope) and isinstance(scope2, IDScope):
        if ids := sorted(set(scope1.ids) & set(scope2.ids)):
            return IDScope(ids=ids)
        return None
    elif isinstance(scope1, IDScopeLowerCase) and isinstance(scope2, IDScopeLowerCase):
        if ids := sorted(set(scope1.ids) & set(scope2.ids)):
            return IDScopeLowerCase(ids=ids)
        return None
    elif isinstance(scope1, SpaceIDScope) and isinstance(scope2, SpaceIDScope):
        if intersection_space_ids := sorted(set(scope1.space_ids) & set(scope2.space_ids)):
            return SpaceIDScope(space_ids=intersection_space_ids)
        return None
    elif isinstance(scope1, AssetRootIDScope) and isinstance(scope2, AssetRootIDScope):
        if intersection_root_ids := sorted(set(scope1.root_ids) & set(scope2.root_ids)):
            return AssetRootIDScope(root_ids=intersection_root_ids)
        return None
    elif isinstance(scope1, CurrentUserScope) and isinstance(scope2, CurrentUserScope):
        return CurrentUserScope()
    elif isinstance(scope1, ExtractionPipelineScope) and isinstance(scope2, ExtractionPipelineScope):
        if intersection_ids := sorted(set(scope1.ids) & set(scope2.ids)):
            return ExtractionPipelineScope(ids=intersection_ids)
        return None
    elif isinstance(scope1, TableScope) and isinstance(scope2, TableScope):
        intersection_tables = set(scope1.dbs_to_tables.keys()) & set(scope2.dbs_to_tables.keys())
        if intersection_tables:
            return TableScope(
                dbs_to_tables={
                    db: sorted(set(scope1.dbs_to_tables[db]) & set(scope2.dbs_to_tables[db]))
                    for db in sorted(intersection_tables)
                }
            )
        return None
    elif isinstance(scope1, InstancesScope) and isinstance(scope2, InstancesScope):
        if intersection_instances := sorted(set(scope1.instances) & set(scope2.instances)):
            return InstancesScope(instances=intersection_instances)
        return None
    elif isinstance(scope1, PartitionScope) and isinstance(scope2, PartitionScope):
        if intersection_partition_ids := sorted(set(scope1.partition_ids) & set(scope2.partition_ids)):
            return PartitionScope(partition_ids=intersection_partition_ids)
        return None
    elif isinstance(scope1, LegacySpaceScope) and isinstance(scope2, LegacySpaceScope):
        if intersection_external_ids := sorted(set(scope1.external_ids) & set(scope2.external_ids)):
            return LegacySpaceScope(external_ids=intersection_external_ids)
        return None
    elif isinstance(scope1, LegacyDataModelScope) and isinstance(scope2, LegacyDataModelScope):
        if intersection_external_ids := sorted(set(scope1.external_ids) & set(scope2.external_ids)):
            return LegacyDataModelScope(external_ids=intersection_external_ids)
        return None
    elif isinstance(scope1, ExperimentsScope) and isinstance(scope2, ExperimentsScope):
        if intersection_experiments := sorted(set(scope1.experiments) & set(scope2.experiments)):
            return ExperimentsScope(experiments=intersection_experiments)
        return None
    elif isinstance(scope1, PostgresGatewayUsersScope) and isinstance(scope2, PostgresGatewayUsersScope):
        if intersection_usernames := sorted(set(scope1.usernames) & set(scope2.usernames)):
            return PostgresGatewayUsersScope(usernames=intersection_usernames)
        return None
    else:
        raise TypeError(f"Unknown scope type {type(scope1)!r}")


def scope_union(scope1: T_Scope, scope2: T_Scope | None) -> Capability.Scope:
    """
    Returns the union of two scopes.
    If both scopes are None, returns None.
    If one scope is None, returns the other scope.
    If both scopes are the same, returns the union of the two scopes.
    If scopes are different types, raise an ValueError.
    """
    if scope2 is None:
        return scope1
    if type(scope1) is not type(scope2):
        raise ValueError("Cannot union scopes of different types")
    if isinstance(scope1, AllScope) or isinstance(scope2, AllScope):
        return AllScope()
    elif isinstance(scope1, AppConfigScope) and isinstance(scope2, AppConfigScope):
        return AppConfigScope(apps=sorted(set(scope1.apps) | set(scope2.apps)))
    elif isinstance(scope1, DataSetScope) and isinstance(scope2, DataSetScope):
        return DataSetScope(ids=sorted(set(scope1.ids) | set(scope2.ids)))
    elif isinstance(scope1, IDScope) and isinstance(scope2, IDScope):
        return IDScope(ids=sorted(set(scope1.ids) | set(scope2.ids)))
    elif isinstance(scope1, IDScopeLowerCase) and isinstance(scope2, IDScopeLowerCase):
        return IDScopeLowerCase(ids=sorted(set(scope1.ids) | set(scope2.ids)))
    elif isinstance(scope1, SpaceIDScope) and isinstance(scope2, SpaceIDScope):
        return SpaceIDScope(space_ids=sorted(set(scope1.space_ids) | set(scope2.space_ids)))
    elif isinstance(scope1, AssetRootIDScope) and isinstance(scope2, AssetRootIDScope):
        return AssetRootIDScope(root_ids=sorted(set(scope1.root_ids) | set(scope2.root_ids)))
    elif isinstance(scope1, CurrentUserScope) and isinstance(scope2, CurrentUserScope):
        return CurrentUserScope()
    elif isinstance(scope1, ExtractionPipelineScope) and isinstance(scope2, ExtractionPipelineScope):
        return ExtractionPipelineScope(ids=sorted(set(scope1.ids) | set(scope2.ids)))
    elif isinstance(scope1, TableScope) and isinstance(scope2, TableScope):
        union_tables = set(scope1.dbs_to_tables.keys()) | set(scope2.dbs_to_tables.keys())
        return TableScope(
            dbs_to_tables={
                db: sorted(set(scope1.dbs_to_tables.get(db, [])) | set(scope2.dbs_to_tables.get(db, [])))
                for db in sorted(union_tables)
            }
        )
    elif isinstance(scope1, InstancesScope) and isinstance(scope2, InstancesScope):
        return InstancesScope(
            instances=sorted(set(scope1.instances) | set(scope2.instances)),
        )
    elif isinstance(scope1, PartitionScope) and isinstance(scope2, PartitionScope):
        return PartitionScope(
            partition_ids=sorted(set(scope1.partition_ids) | set(scope2.partition_ids)),
        )
    elif isinstance(scope1, LegacySpaceScope) and isinstance(scope2, LegacySpaceScope):
        return LegacySpaceScope(
            external_ids=sorted(set(scope1.external_ids) | set(scope2.external_ids)),
        )
    elif isinstance(scope1, LegacyDataModelScope) and isinstance(scope2, LegacyDataModelScope):
        return LegacyDataModelScope(
            external_ids=sorted(set(scope1.external_ids) | set(scope2.external_ids)),
        )
    elif isinstance(scope1, ExperimentsScope) and isinstance(scope2, ExperimentsScope):
        return ExperimentsScope(
            experiments=sorted(set(scope1.experiments) | set(scope2.experiments)),
        )
    elif isinstance(scope1, PostgresGatewayUsersScope) and isinstance(scope2, PostgresGatewayUsersScope):
        return PostgresGatewayUsersScope(
            usernames=sorted(set(scope1.usernames) | set(scope2.usernames)),
        )
    else:
        raise TypeError(f"Unknown scope type {type(scope1)!r}")



import json
from collections.abc import Sequence

import pytest
import requests
import responses
from _pytest.mark import ParameterSet
from cognite.client.data_classes import Event, FileMetadata
from cognite.client.data_classes._base import CogniteResource
from cognite.client.data_classes.capabilities import Capability, EventsAcl, FilesAcl

from cognite_toolkit._cdf_tk.client import ToolkitClient, ToolkitClientConfig
from cognite_toolkit._cdf_tk.exceptions import AuthorizationError
from tests.test_unit.utils import FakeCogniteResourceGenerator


@pytest.mark.usefixtures("disable_gzip", "disable_pypi_check")
class TestLookup:
    CASES: Sequence[ParameterSet] = [
        pytest.param("events", Event, EventsAcl, id="events"),
        pytest.param("files", FileMetadata, FilesAcl, id="files"),
    ]

    @pytest.mark.parametrize("endpoint,resource_cls,cap_cls", CASES)
    def test_lookup_id(
        self,
        endpoint: str,
        resource_cls: type[CogniteResource],
        cap_cls: type[Capability],
        toolkit_config: ToolkitClientConfig,
        rsps: responses.RequestsMock,
    ) -> None:
        resources = self._create_resources(resource_cls, N=5)
        client = self._create_client_mock(endpoint, resources, rsps, toolkit_config)
        lookup_api = getattr(client.lookup, endpoint)

        multiple_external_ids = lookup_api.external_id(id=[resource.id for resource in resources])
        assert multiple_external_ids == [resource.external_id for resource in resources]

        single_external_di = lookup_api.external_id(id=resources[0].id)
        assert single_external_di == resources[0].external_id

        multiple_ids = lookup_api.id(external_id=[resource.external_id for resource in resources])
        assert multiple_ids == [resource.id for resource in resources]

        single_id = lookup_api.id(external_id=resources[0].external_id)
        assert single_id == resources[0].id

    @pytest.mark.parametrize("endpoint,resource_cls,cap_cls", CASES)
    def test_lookup_external_id(
        self,
        endpoint: str,
        resource_cls: type[CogniteResource],
        cap_cls: type[Capability],
        toolkit_config: ToolkitClientConfig,
        rsps: responses.RequestsMock,
    ) -> None:
        resources = self._create_resources(resource_cls, N=5)
        client = self._create_client_mock(endpoint, resources, rsps, toolkit_config)
        lookup_api = getattr(client.lookup, endpoint)

        # We are testing both orders as the lookup caches the results.
        single_id = lookup_api.id(external_id=resources[0].external_id)
        assert single_id == resources[0].id

        multiple_ids = lookup_api.id(external_id=[resource.external_id for resource in resources])
        assert multiple_ids == [resource.id for resource in resources]

        single_external_di = lookup_api.external_id(id=resources[0].id)
        assert single_external_di == resources[0].external_id

        multiple_external_ids = lookup_api.external_id(id=[resource.id for resource in resources])
        assert multiple_external_ids == [resource.external_id for resource in resources]

    @pytest.mark.parametrize("endpoint,resource_cls,cap_cls", CASES)
    def test_missing_access(
        self,
        endpoint: str,
        resource_cls: type[CogniteResource],
        cap_cls: type[Capability],
        toolkit_config: ToolkitClientConfig,
        rsps: responses.RequestsMock,
    ) -> None:
        rsps.post(
            toolkit_config.create_api_url(f"{endpoint}/byids"),
            status=403,
        )
        rsps.add(
            responses.GET,
            f"{toolkit_config.base_url}/api/v1/token/inspect",
            json={
                "subject": "123",
                "projects": [],
                "capabilities": [],
            },
        )
        client = ToolkitClient(config=toolkit_config)
        lookup_api = getattr(client.lookup, endpoint)

        with pytest.raises(AuthorizationError) as exc_info:
            lookup_api.id(external_id="non-existing")

        assert cap_cls.__name__ in str(exc_info.value)

    def test_lookup_external_id_missing(
        self,
        toolkit_config: ToolkitClientConfig,
        rsps: responses.RequestsMock,
    ) -> None:
        config = toolkit_config
        rsps.post(
            config.create_api_url("assets/byids"),
            status=200,
            json={"items": [{"id": 1, "externalId": "ext-1"}]},
        )
        client = ToolkitClient(config=config)
        result = client.lookup.assets.external_id([1, 2, 3])
        assert result == ["ext-1"]
        result2 = client.lookup.assets.external_id([1, 2, 3])
        assert result2 == ["ext-1"]

        assert len(rsps.calls) == 1  # Cached result used for second call

    @staticmethod
    def _create_resources(resource_cls: type[CogniteResource], N: int) -> list[CogniteResource]:
        generator = FakeCogniteResourceGenerator(seed=42)
        resources = [generator.create_instance(resource_cls) for _ in range(N)]
        for i, resource in enumerate(resources):
            resource.id = i + 1000
            resource.external_id = f"ext-{i + 1000}"
        return resources

    @staticmethod
    def _create_client_mock(
        endpoint: str, resources: list[CogniteResource], rsps: responses.RequestsMock, config: ToolkitClientConfig
    ) -> ToolkitClient:
        resource_by_id = {resource.id: resource for resource in resources}
        resource_by_external_id = {resource.external_id: resource for resource in resources}

        def retrieve_multiple_callback(request: requests.PreparedRequest) -> tuple[int, dict[str, str], str]:
            items = json.loads(request.body)["items"]
            response_items = []
            for item in items:
                if "id" in item and (resource := resource_by_id.get(item["id"])) is not None:
                    response_items.append(resource.dump(camel_case=True))
                elif "externalId" in item and (resource := resource_by_external_id.get(item["externalId"])) is not None:
                    response_items.append(resource.dump(camel_case=True))
            # Return format: (status code, headers, body)
            return 200, {}, json.dumps({"items": response_items})

        rsps.add_callback(
            responses.POST, config.create_api_url(f"{endpoint}/byids"), callback=retrieve_multiple_callback
        )
        client = ToolkitClient(config)
        return client



import threading
from collections import Counter
from typing import Generic, Literal, TypeAlias, get_args, overload

from cognite_toolkit._cdf_tk.utils import humanize_collection
from cognite_toolkit._cdf_tk.utils.useful_types import T_ID

Status: TypeAlias = Literal["success", "failed", "aborted", "pending"]

AVAILABLE_STATUS = get_args(Status)


class ProgressTracker(Generic[T_ID]):
    """A thread-safe progress tracker for multiple items and steps.

    For example, when migrating asset-centric Assets to CogniteAsset, we would like to track the progress of each Asset
    through the steps of the migration. This is such that we can report back to the user which Assets failed, where
    they failed, or if they succeeded.

    Args:
        steps (list[str]): A list of unique step names to track progress for.

    Examples:
        >>> tracker = ProgressTracker(steps=["downloadAsset", "convert", "uploadCogniteAssset"])
        >>> tracker.set_progress(item_id=123, step="downloadAsset", status="success")
        >>> tracker.set_progress(item_id=123, step="convert", status="failed")
        >>> tracker.get_progress(item_id=123)
        {'downloadAsset': 'success', 'convert': 'failed', 'uploadCogniteAssset': 'aborted'}

    """

    def __init__(self, steps: list[str]) -> None:
        if len(steps) != len(set(steps)):
            raise ValueError("Input `steps` must be unique.")
        self._steps = list(steps)
        self._step_to_idx: dict[str, int] = {step: i for i, step in enumerate(self._steps)}
        self._progress: dict[T_ID, dict[str, Status]] = {}
        self._lock = threading.Lock()

    def _init_item(self, item_id: T_ID) -> None:
        if item_id not in self._progress:
            self._progress[item_id] = {step: "pending" for step in self._steps}

    def set_progress(self, item_id: T_ID, step: str, status: Status) -> None:
        """Set the progress of a specific step for a given item."""
        if status not in AVAILABLE_STATUS:
            raise ValueError(f"Status must be one of {humanize_collection(AVAILABLE_STATUS)}")
        with self._lock:
            self._init_item(item_id)
            try:
                idx = self._step_to_idx[step]
            except KeyError as e:
                raise ValueError(f"Step '{step}' not found in steps {humanize_collection(self._steps)}") from e
            self._progress[item_id][step] = status
            if status == "failed":
                for s in self._steps[idx + 1 :]:
                    self._progress[item_id][s] = "aborted"

    @overload
    def get_progress(self, item_id: T_ID) -> dict[str, Status]: ...

    @overload
    def get_progress(self, item_id: T_ID, step: str) -> Status: ...

    def get_progress(self, item_id: T_ID, step: str | None = None) -> dict[str, Status] | Status:
        """Get the progress of all steps or a specific step for a given item."""
        with self._lock:
            self._init_item(item_id)
            if step is None:
                return self._progress[item_id].copy()
            if step not in self._step_to_idx:
                raise ValueError(f"Step '{step}' not found in steps {humanize_collection(self._steps)}")
            return self._progress[item_id][step]

    def result(self) -> dict[T_ID, dict[str, Status]]:
        """Get the progress of all items."""
        with self._lock:
            return {item_id: progress.copy() for item_id, progress in self._progress.items()}

    def aggregate(self) -> dict[tuple[str, Status], int]:
        """Aggregate the progress across all items and steps.

        Returns:
            dict[tuple[str, Status], int]: A dictionary with keys as (step, status) tuples and values as counts.
        """
        with self._lock:
            return Counter((step, status) for progress in self._progress.values() for step, status in progress.items())



from pathlib import Path

from cognite.client.data_classes import (
    Asset,
    AssetList,
    CountAggregate,
    DataSet,
    Event,
    EventList,
    FileMetadata,
    FileMetadataList,
    GeoLocation,
    Geometry,
    Label,
    LabelDefinition,
    LabelDefinitionList,
    TimeSeries,
    TimeSeriesList,
    TransformationPreviewResult,
)

from cognite_toolkit._cdf_tk.client.testing import monkeypatch_toolkit_client
from cognite_toolkit._cdf_tk.commands import DumpDataCommand
from cognite_toolkit._cdf_tk.commands.dump_data import AssetFinder, EventFinder, FileMetadataFinder, TimeSeriesFinder
from cognite_toolkit._cdf_tk.utils.file import read_yaml_file


class TestDumpData:
    def test_dump_assets(self, tmp_path: Path) -> None:
        my_label = LabelDefinition(external_id="label1", name="Label 1")
        dataset = DataSet(external_id="my_dataset", name="My Dataset", id=123)
        my_asset = Asset(
            external_id="my_asset",
            name="My Asset",
            parent_external_id=None,
            description="This is my asset",
            data_set_id=dataset.id,
            metadata={"key": "value"},
            source="MySource",
            labels=[my_label.external_id],
            geo_location=GeoLocation(
                type="Feature",
                geometry=Geometry(type="LineString", coordinates=[[1.0, 1.0], [2.0, 2.0], [3.0, 3.0]]),
                properties={},
            ),
        )
        cmd = DumpDataCommand(skip_tracking=False, print_warning=False)
        output_dir = tmp_path / "asset_dump"
        csv_dir = output_dir / "csv"
        parquet_dir = output_dir / "parquet"
        with monkeypatch_toolkit_client() as client:
            client.assets.return_value = [AssetList([my_asset])]
            client.assets.aggregate_count.return_value = 1
            client.assets.retrieve.return_value = Asset(external_id="rootAsset")
            client.labels.retrieve.return_value = LabelDefinitionList([my_label])
            client.data_sets.retrieve_multiple.return_value = [dataset]
            client.lookup.assets.external_id.return_value = "rootAsset"
            client.lookup.data_sets.external_id.return_value = dataset.external_id
            client.transformations.preview.return_value = TransformationPreviewResult(
                None, [{"key": "key", "key_count": 1}]
            )

            cmd.dump_table(
                AssetFinder(client, ["rootAsset"], []),
                csv_dir,
                clean=True,
                limit=None,
                format_="csv",
                verbose=False,
            )
            cmd.dump_table(
                AssetFinder(client, [], [dataset.external_id]),
                parquet_dir,
                clean=True,
                limit=None,
                format_="parquet",
                verbose=False,
            )

        output_csvs = list(csv_dir.rglob("*.csv"))
        assert len(output_csvs) == 1
        output_csv = output_csvs[0]
        assert output_csv.read_text().splitlines() == [
            "externalId,name,parentExternalId,description,dataSetExternalId,source,labels,geoLocation,metadata.key",
            "my_asset,My Asset,,This is my asset,my_dataset,MySource,['label1'],"
            "\"{'type': 'Feature', 'geometry': {'type': 'LineString', 'coordinates': [[1.0, 1.0], [2.0, 2.0], "
            "[3.0, 3.0]]}, 'properties': {}}\""
            ",value",
        ]

        dataset_yamls = list(csv_dir.rglob("*DataSet.yaml"))
        assert len(dataset_yamls) == 1
        dataset_yaml = dataset_yamls[0]
        assert read_yaml_file(dataset_yaml) == [dataset.as_write().dump()]

        label_yamls = list(csv_dir.rglob("*Label.yaml"))
        assert len(label_yamls) == 1
        label_yaml = label_yamls[0]
        assert read_yaml_file(label_yaml) == [my_label.as_write().dump()]

        parquet_files = list(parquet_dir.rglob("*.parquet"))
        assert len(parquet_files) == 1

    def test_dump_filemetadata(self, tmp_path: Path) -> None:
        my_label = LabelDefinition(external_id="label1", name="Label 1")
        dataset = DataSet(external_id="my_dataset", name="My Dataset", id=123)
        my_file = FileMetadata(
            external_id="my_file",
            name="My File",
            data_set_id=dataset.id,
            metadata={"key": "value"},
            source="MySource",
            labels=[Label(external_id=my_label.external_id)],
            asset_ids=[1],
            mime_type="application/octet-stream",
            directory="my_directory",
            geo_location=GeoLocation(
                type="Feature",
                geometry=Geometry(type="LineString", coordinates=[[1.0, 1.0], [2.0, 2.0], [3.0, 3.0]]),
                properties={},
            ),
        )
        my_other_file = FileMetadata(
            "my_other_file_タシ",
            name="My Other File",
        )
        cmd = DumpDataCommand(skip_tracking=False, print_warning=False)
        output_dir = tmp_path / "file_dump"
        with monkeypatch_toolkit_client() as client:
            client.files.return_value = [FileMetadataList([my_file, my_other_file])]
            client.files.aggregate.return_value = [CountAggregate(count=2)]
            client.labels.retrieve.return_value = LabelDefinitionList([my_label])
            client.data_sets.retrieve_multiple.return_value = [dataset]
            client.lookup.assets.external_id.return_value = "rootAsset"
            client.lookup.data_sets.external_id.return_value = dataset.external_id
            client.transformations.preview.return_value = TransformationPreviewResult(
                None, [{"key": "key", "key_count": 1}]
            )

            cmd.dump_table(
                FileMetadataFinder(client, [], ["my_dataset"]),
                output_dir,
                clean=True,
                limit=None,
                format_="csv",
                verbose=False,
            )

        output_csvs = list(output_dir.rglob("*.csv"))
        assert len(output_csvs) == 1
        output_csv = output_csvs[0]
        assert output_csv.read_text(encoding="utf-8").splitlines() == [
            "externalId,name,directory,source,mimeType,assetExternalIds,dataSetExternalId,sourceCreatedTime,"
            "sourceModifiedTime,securityCategories,labels,geoLocation,metadata.key",
            "my_file,My File,my_directory,MySource,application/octet-stream,rootAsset,my_dataset,,,,['label1'],"
            "\"{'type': 'Feature', 'geometry': {'type': 'LineString', 'coordinates': [[1.0, 1.0], [2.0, 2.0], "
            "[3.0, 3.0]]}, 'properties': {}}\""
            ",value",
            "my_other_file_タシ,My Other File,,,,,,,,,,,",
        ]

        dataset_yamls = list(output_dir.rglob("*DataSet.yaml"))
        assert len(dataset_yamls) == 1
        dataset_yaml = dataset_yamls[0]
        assert read_yaml_file(dataset_yaml) == [dataset.as_write().dump()]

        label_yamls = list(output_dir.rglob("*Label.yaml"))
        assert len(label_yamls) == 1
        label_yaml = label_yamls[0]
        assert read_yaml_file(label_yaml) == [my_label.as_write().dump()]

    def test_dump_timeseries(self, tmp_path: Path) -> None:
        dataset = DataSet(external_id="my_dataset", name="My Dataset", id=123)
        my_timeseries = TimeSeries(
            external_id="my_timeseries",
            name="My TimeSeries",
            description="This is my timeseries",
            metadata={"key": "value"},
            is_string=False,
            is_step=False,
            data_set_id=dataset.id,
        )
        cmd = DumpDataCommand(skip_tracking=False, print_warning=False)
        output_dir = tmp_path / "timeseries_dump"
        with monkeypatch_toolkit_client() as client:
            client.time_series.return_value = [TimeSeriesList([my_timeseries])]
            client.time_series.aggregate_count.return_value = 1
            client.data_sets.retrieve_multiple.return_value = [dataset]
            client.lookup.data_sets.external_id.return_value = dataset.external_id
            client.transformations.preview.return_value = TransformationPreviewResult(
                None, [{"key": "key", "key_count": 1}]
            )

            cmd.dump_table(
                TimeSeriesFinder(client, [], [dataset.external_id]),
                output_dir,
                clean=True,
                limit=None,
                format_="csv",
                verbose=False,
            )

        output_csv = next(output_dir.rglob("*.csv"))
        assert output_csv.read_text().splitlines() == [
            "externalId,name,isString,unit,unitExternalId,assetExternalId,isStep,description,dataSetExternalId,securityCategories,metadata.key",
            "my_timeseries,My TimeSeries,False,,,,False,This is my timeseries,my_dataset,,value",
        ]

        dataset_yaml = next(output_dir.rglob("*DataSet.yaml"))
        assert read_yaml_file(dataset_yaml) == [dataset.as_write().dump()]

    def test_dump_timeseries_in_parallel(self, tmp_path: Path) -> None:
        my_timeseries_list = [
            TimeSeries(
                external_id=f"my_timeseries_{i}",
                name=f"My TimeSeries {i}",
                is_string=False,
                is_step=False,
            )
            for i in range(10)
        ]
        cmd = DumpDataCommand(skip_tracking=False, print_warning=False)
        output_dir = tmp_path / "timeseries_dump"
        with monkeypatch_toolkit_client() as client:
            client.time_series.return_value = [TimeSeriesList([ts]) for ts in my_timeseries_list]
            # The iteration count is the number of timeseries / 1000
            client.time_series.aggregate_count.return_value = 100_000
            client.transformations.preview.return_value = TransformationPreviewResult(None, [])

            cmd.dump_table(
                TimeSeriesFinder(client, [], ["doesn't matter"]),
                output_dir,
                clean=True,
                limit=None,
                format_="csv",
                verbose=False,
                parallel_threshold=2,
                max_queue_size=4,
            )

        output_csvs = list(output_dir.rglob("*.csv"))
        assert len(output_csvs) == 1
        output_csv = output_csvs[0]
        assert output_csv.read_text().splitlines() == [
            "externalId,name,isString,unit,unitExternalId,assetExternalId,isStep,description,dataSetExternalId,securityCategories",
        ] + [f"my_timeseries_{i},My TimeSeries {i},False,,,,False,,," for i in range(10)]

    def test_dump_events(self, tmp_path: Path) -> None:
        dataset = DataSet(external_id="my_dataset", name="My Dataset", id=123)
        my_event = Event(
            external_id="my_event",
            description="This is my event",
            start_time=1,
            end_time=2,
            type="my_type",
            subtype="my_subtype",
            metadata={"key": "value", "key2": "value2"},
            data_set_id=dataset.id,
            source="my_source",
        )
        cmd = DumpDataCommand(skip_tracking=False, print_warning=False)
        output_dir = tmp_path / "event_dump"
        with monkeypatch_toolkit_client() as client:
            client.events.return_value = [EventList([my_event])]
            client.events.aggregate_count.return_value = 1
            client.data_sets.retrieve_multiple.return_value = [dataset]
            client.lookup.data_sets.external_id.return_value = dataset.external_id
            client.transformations.preview.return_value = TransformationPreviewResult(
                None, [{"key": "key", "key_count": 1}, {"key": "key2", "key_count": 1}]
            )

            cmd.dump_table(
                EventFinder(client, [], [dataset.external_id]),
                output_dir,
                clean=True,
                limit=None,
                format_="csv",
                verbose=False,
            )

        output_csvs = list(output_dir.rglob("*.csv"))
        assert len(output_csvs) == 1
        output_csv = output_csvs[0]
        assert output_csv.read_text().splitlines() == [
            "externalId,dataSetExternalId,startTime,endTime,type,subtype,description,assetExternalIds,source,metadata.key,metadata.key2",
            "my_event,my_dataset,1,2,my_type,my_subtype,This is my event,,my_source,value,value2",
        ]

        dataset_yamls = list(output_dir.rglob("*DataSet.yaml"))
        assert len(dataset_yamls) == 1
        dataset_yaml = dataset_yamls[0]
        assert read_yaml_file(dataset_yaml) == [dataset.as_write().dump()]



from __future__ import annotations

import pytest

from cognite_toolkit._cdf_tk.cdf_toml import CDFToml
from cognite_toolkit._cdf_tk.plugins import Plugin, Plugins
from tests.data import CDF_TOML_DATA


@pytest.fixture(autouse=True)
def non_singleton_cdf_toml() -> None:
    _ = CDFToml.load(CDF_TOML_DATA, use_singleton=True)


@pytest.mark.skip("This test is not working because we need to mock the CDFToml.load method first")
class TestPlugins:
    def test_list_all(self) -> None:
        plugins = Plugins.list()
        assert "run" in plugins
        assert "pull" in plugins
        assert "dump" in plugins

    def test_plugin_disabled(self) -> None:
        assert not Plugin.is_enabled(Plugins.dump)

    def test_plugin_enabled(self) -> None:
        assert Plugin.is_enabled(Plugins.run)



from pathlib import Path

import pytest

from cognite_toolkit._cdf_tk.resource_classes.robotics.capability import RobotCapabilityYAML
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml_pydantic
from tests.test_unit.utils import find_resources


class TestRobotCapabilityYAML:
    @pytest.mark.parametrize("data", list(find_resources("RobotCapability")))
    def test_load_valid_robot_capability_file(self, data: dict[str, object]) -> None:
        loaded = RobotCapabilityYAML.model_validate(data)
        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize(
        "data",
        [
            pytest.param(
                {
                    "name": "PTZ Camera",
                    "externalId": "ptz_camera_001",
                    "method": "capture_ptz_image",
                    "description": "Pan-Tilt-Zoom camera capability for capturing images",
                    "inputSchema": {
                        "type": "object",
                        "properties": {
                            "pan": {"type": "number", "minimum": -180, "maximum": 180},
                            "tilt": {"type": "number", "minimum": -90, "maximum": 90},
                            "zoom": {"type": "number", "minimum": 1, "maximum": 10},
                        },
                    },
                    "dataHandlingSchema": {
                        "uploadLocation": "cdf_files",
                        "fileFormat": "jpeg",
                        "metadata": {"captureType": "ptz"},
                    },
                },
                id="full-ptz-camera-capability",
            ),
            pytest.param(
                {"name": "Docking Station", "externalId": "dock_001", "method": "return_to_dock"},
                id="minimal-robot-capability",
            ),
        ],
    )
    def test_load_valid_robot_capability(self, data: dict[str, object]) -> None:
        loaded = RobotCapabilityYAML.model_validate(data)
        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize(
        "data, expected_errors",
        [
            pytest.param(
                {"name": "Missing External ID", "method": "some_method"},
                {"Missing required field: 'externalId'"},
                id="missing-required-field",
            ),
            pytest.param(
                {"externalId": "cap_001", "name": "Test Capability", "method": "test_method", "unknownField": "value"},
                {"Unused field: 'unknownField'"},
                id="unknown-field-present",
            ),
        ],
    )
    def test_invalid_robot_capability_error_messages(self, data: dict | list, expected_errors: set[str]) -> None:
        warning_list = validate_resource_yaml_pydantic(data, RobotCapabilityYAML, Path("some_file.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)
        assert set(format_warning.errors) == expected_errors



import sys
from collections.abc import Sequence
from typing import Any, cast

from cognite.client import CogniteClient
from cognite.client.data_classes.data_modeling import NodeId
from cognite.client.data_classes.time_series import TimeSeries, TimeSeriesList

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class ExtendedTimeSeries(TimeSeries):
    """Extended TimeSeries with pending instance ID support.

    Args:
        id (int | None): A server-generated ID for the object.
        external_id (str | None): The externally supplied ID for the time series.
        instance_id (NodeId | None): The Instance ID for the time series. (Only applicable for time series created in DMS)
        name (str | None): The display short name of the time series.
        is_string (bool | None): Whether the time series is string valued or not.
        metadata (dict[str, str] | None): Custom, application-specific metadata. String key -> String value. Limits: Maximum length of key is 32 bytes, value 512 bytes, up to 16 key-value pairs.
        unit (str | None): The physical unit of the time series.
        unit_external_id (str | None): The physical unit of the time series (reference to unit catalog). Only available for numeric time series.
        asset_id (int | None): Asset ID of equipment linked to this time series.
        is_step (bool | None): Whether the time series is a step series or not.
        description (str | None): Description of the time series.
        security_categories (Sequence[int] | None): The required security categories to access this time series.
        data_set_id (int | None): The dataSet ID for the item.
        created_time (int | None): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        last_updated_time (int | None): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        pending_instance_id (NodeId | None): The pending instance ID for the time series. This is used in the migration of times series to CogniteTimeSeries
            to specify which instance this time series should be linked with by the TimeSeries syncer service.
        legacy_name (str | None): This field is not used by the API and will be removed October 2024.
        cognite_client (CogniteClient | None): The client to associate with this object.
    """

    def __init__(
        self,
        id: int | None = None,
        external_id: str | None = None,
        instance_id: NodeId | None = None,
        name: str | None = None,
        is_string: bool | None = None,
        metadata: dict[str, str] | None = None,
        unit: str | None = None,
        unit_external_id: str | None = None,
        asset_id: int | None = None,
        is_step: bool | None = None,
        description: str | None = None,
        security_categories: Sequence[int] | None = None,
        data_set_id: int | None = None,
        created_time: int | None = None,
        last_updated_time: int | None = None,
        pending_instance_id: NodeId | None = None,
        legacy_name: str | None = None,
        cognite_client: CogniteClient | None = None,
    ) -> None:
        super().__init__(
            external_id=external_id,
            instance_id=instance_id,
            name=name,
            is_string=is_string,
            metadata=metadata,
            unit=unit,
            unit_external_id=unit_external_id,
            asset_id=asset_id,
            is_step=is_step,
            description=description,
            security_categories=security_categories,
            data_set_id=data_set_id,
            legacy_name=legacy_name,
        )
        # id/created_time/last_updated_time are required when using the class to read,
        # but don't make sense passing in when creating a new object. So in order to make the typing
        # correct here (i.e. int and not Optional[int]), we force the type to be int rather than
        # Optional[int].
        # TODO: In the next major version we can make these properties required in the constructor
        self.id: int = id  # type: ignore
        self.created_time: int = created_time  # type: ignore
        self.last_updated_time: int = last_updated_time  # type: ignore
        self._cognite_client = cast("CogniteClient", cognite_client)
        self.pending_instance_id = pending_instance_id

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        instance = super()._load(resource, cognite_client)
        if isinstance(instance.pending_instance_id, dict):
            instance.pending_instance_id = NodeId.load(instance.pending_instance_id)
        return instance

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        """Dump the object to a dictionary"""
        output = super().dump(camel_case=camel_case)
        if self.pending_instance_id is not None:
            output["pendingInstanceId" if camel_case else "pending_instance_id"] = self.pending_instance_id.dump(
                camel_case=camel_case, include_instance_type=False
            )
        return output


class ExtendedTimeSeriesList(TimeSeriesList):
    _RESOURCE = ExtendedTimeSeries  # type: ignore[assignment]



from __future__ import annotations

from dataclasses import dataclass
from typing import Literal

from cognite.client.data_classes import Group
from cognite.client.data_classes._base import CogniteResource, CogniteResourceList
from cognite.client.data_classes.capabilities import Capability, UnknownAcl


@dataclass
class AuthGroupCalls:
    name: str
    calls: list[Group]

    @property
    def last_created_capabilities(self) -> set[str]:
        return {self._get_capability_name(c) for c in self.calls[-1].capabilities}

    @property
    def capabilities_all_calls(self) -> set[str]:
        return {self._get_capability_name(c) for call in self.calls for c in call.capabilities}

    def _get_capability_name(self, capability: Capability) -> str:
        return capability.capability_name if isinstance(capability, UnknownAcl) else capability._capability_name


@dataclass
class Method:
    """Represent a method in the CogniteClient that should be mocked

    Args:
        api_class_method: The name of the method in the CogniteClient, for example, 'create', 'insert_dataframe'
        mock_class_method: The name of the method in the ApprovalCogniteClient, for example, 'create', 'insert_dataframe'

    The available mock methods you can see inside
    * ApprovalCogniteClient._create_create_method,
    * ApprovalCogniteClient._create_delete_method,
    * ApprovalCogniteClient._create_retrieve_method

    """

    api_class_method: str
    mock_class_method: str


@dataclass
class APIResource:
    """This is used to define the resources that should be mocked in the ApprovalCogniteClient

    Args:
        api_name: The name of the resource in the CogniteClient, for example, 'time_series', 'data_modeling.views'
        resource_cls: The resource class for the API
        list_cls: The list resource API class
        methods: The methods that should be mocked
        _write_cls: The write resource class for the API. For example, the writing class for 'data_modeling.views' is 'ViewApply'
        _write_list_cls: The write list class in the CogniteClient

    """

    api_name: str
    resource_cls: type[CogniteResource]
    list_cls: type[CogniteResourceList] | type[list]
    methods: dict[Literal["create", "delete", "retrieve"], list[Method]]

    _write_cls: type[CogniteResource] | None = None
    _write_list_cls: type[CogniteResourceList] | None = None

    @property
    def write_cls(self) -> type[CogniteResource]:
        return self._write_cls or self.resource_cls

    @property
    def write_list_cls(self) -> type[CogniteResourceList]:
        return self._write_list_cls or self.list_cls



import pytest

from cognite_toolkit._cdf_tk.client import ToolkitClientConfig


class TestToolkitClientConfig:
    def test_create_app_url_raise_on_empty(self, toolkit_config: ToolkitClientConfig) -> None:
        with pytest.raises(ValueError, match="Endpoint must be a non-empty string"):
            toolkit_config.create_app_url("")



import pytest
import responses

from cognite_toolkit._cdf_tk.client import ToolkitClient, ToolkitClientConfig
from cognite_toolkit._cdf_tk.client.data_classes.charts import Chart, ChartList
from cognite_toolkit._cdf_tk.client.data_classes.charts_data import ChartData
from cognite_toolkit._cdf_tk.client.testing import monkeypatch_toolkit_client
from cognite_toolkit._cdf_tk.storageio import ChartIO
from cognite_toolkit._cdf_tk.storageio.selectors import (
    AllChartsSelector,
    ChartOwnerSelector,
    ChartSelector,
)


@pytest.fixture
def twenty_charts() -> ChartList:
    return ChartList(
        [
            Chart(
                external_id=f"chart_{i}",
                visibility="PUBLIC",
                data=ChartData(),
                owner_id="evenOwner" if i % 2 == 0 else "oddOwner",
                created_time=1700000000000,
                last_updated_time=1700000000000,
            )
            for i in range(20)
        ]
    )


class TestChartIO:
    def test_download_chart(self, toolkit_config: ToolkitClientConfig) -> None:
        client = ToolkitClient(config=toolkit_config)
        chart_url = toolkit_config.create_app_url("/storage/charts/charts/list")
        ts_url = toolkit_config.create_api_url("/timeseries/byids")
        selector = AllChartsSelector()
        io = ChartIO(client)

        with responses.RequestsMock() as rsps:
            rsps.add(
                responses.POST,
                chart_url,
                json={
                    "items": [
                        {
                            "externalId": "chart_1",
                            "ownerId": 100,
                            "visibility": "PUBLIC",
                            "createdTime": 1700000000000,
                            "lastUpdatedTime": 1700000000000,
                            "data": {
                                "name": "Chart 1",
                                "timeSeriesCollection": [
                                    {"tsId": 200, "tsExternalId": None},
                                    {"tsId": None, "tsExternalId": "ts_2"},
                                ],
                            },
                        },
                        {
                            "externalId": "chart_2",
                            "ownerId": 101,
                            "visibility": "PUBLIC",
                            "createdTime": 1700000000000,
                            "lastUpdatedTime": 1700000000000,
                            "data": {
                                "name": "Chart 2",
                                "timeSeriesCollection": [
                                    {"tsId": 201, "tsExternalId": None},
                                    {"tsId": None, "tsExternalId": "ts_3"},
                                ],
                            },
                        },
                    ]
                },
                status=200,
            )
            rsps.add(
                responses.POST,
                ts_url,
                json={"items": [{"id": 200, "externalId": "ts_1"}, {"id": 201, "externalId": "ts_4"}]},
                status=200,
            )
            assert io.count(selector) is None, (
                "Count should be None since CDF does not provide a way to get the count of charts up front."
            )
            charts_iterator = io.stream_data(selector=selector)
            json_iterator = (io.data_to_json_chunk(chunk.items) for chunk in charts_iterator)
            chart_data = [io.json_chunk_to_data([("id", item) for item in chunk]) for chunk in json_iterator]

            assert len(chart_data) == 1
            chart_list = chart_data[0]
            assert len(chart_list) == 2
            first = chart_list[0]
            assert first.item.data.time_series_collection[0].ts_external_id == "ts_1"
            assert first.item.data.time_series_collection[1].ts_external_id == "ts_2"
            second = chart_list[1]
            assert second.item.data.time_series_collection[0].ts_external_id == "ts_4"
            assert second.item.data.time_series_collection[1].ts_external_id == "ts_3"

    @pytest.mark.parametrize(
        "limit,selector,expected_external_ids",
        [
            pytest.param(None, AllChartsSelector(), [f"chart_{i}" for i in range(20)], id="all charts no limit"),
            pytest.param(5, AllChartsSelector(), [f"chart_{i}" for i in range(5)], id="all charts with limit"),
            pytest.param(
                10,
                AllChartsSelector(),
                [f"chart_{i}" for i in range(10)],
                id="all charts with limit 10 divisible by chunk size",
            ),
            pytest.param(
                None,
                ChartOwnerSelector(owner_id="evenOwner"),
                [f"chart_{i}" for i in range(0, 20, 2)],
                id="even owner no limit",
            ),
            pytest.param(
                3,
                ChartOwnerSelector(owner_id="evenOwner"),
                [f"chart_{i}" for i in range(0, 6, 2)],
                id="even owner with limit",
            ),
        ],
    )
    def test_download_iterable(
        self, limit: int | None, selector: ChartSelector, expected_external_ids: list[str], twenty_charts: ChartList
    ) -> None:
        with monkeypatch_toolkit_client() as client:
            client.charts.list.return_value = twenty_charts
            io = ChartIO(client)
            chunks = list(io.stream_data(selector=selector, limit=limit))
            all_charts = ChartList([])
            for chunk in chunks:
                all_charts.extend(chunk.items)
            assert [chart.external_id for chart in all_charts] == expected_external_ids



import sys
from abc import abstractmethod
from collections.abc import ItemsView, Iterator, KeysView, Mapping, ValuesView
from dataclasses import dataclass
from types import MappingProxyType
from typing import Any, ClassVar, cast

from cognite.client import CogniteClient
from cognite.client.data_classes._base import (
    CogniteObject,
    CogniteResource,
    WriteableCogniteResource,
    WriteableCogniteResourceList,
)

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


@dataclass(frozen=True)
class RawDatabase(WriteableCogniteResource):
    db_name: str

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(db_name=resource["dbName"])

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {"dbName" if camel_case else "db_name": self.db_name}

    def as_write(self) -> Self:
        return self


@dataclass(frozen=True)
class RawTable(WriteableCogniteResource):
    db_name: str
    table_name: str

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(db_name=resource["dbName"], table_name=resource["tableName"])

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {
            "dbName" if camel_case else "db_name": self.db_name,
            "tableName" if camel_case else "table_name": self.table_name,
        }

    def as_write(self) -> Self:
        return self

    def __str__(self) -> str:
        return f"{self.db_name}.{self.table_name}"


class RawDatabaseList(WriteableCogniteResourceList[RawDatabase, RawDatabase]):
    _RESOURCE = RawDatabase

    def as_write(self) -> Self:
        return self


class RawTableList(WriteableCogniteResourceList[RawTable, RawTable]):
    _RESOURCE = RawTable

    def as_write(self) -> Self:
        return self


@dataclass
class RawProfileColumn(CogniteObject):
    _type: ClassVar[str]
    count: int
    null_count: int

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        for column_type, column_class in _RAW_PROFILE_COLUMN_BY_TYPE.items():
            if column_type in resource:
                return cast(Self, column_class._load_column(resource))
        return cast(Self, UnknownTypeProfileColumn._load_column(resource))

    @classmethod
    @abstractmethod
    def _load_column(cls, resource: dict[str, object]) -> Self:
        """Load a specific type of RawProfileColumn based on the resource data."""
        raise NotImplementedError("Subclasses must implement this method.")

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {
            "count": self.count,
            "nullCount" if camel_case else "null_count": self.null_count,
        }


@dataclass
class StringProfile(CogniteObject):
    length_range: tuple[int, int]
    distinct_count: int
    length_histogram: tuple[tuple[int, int], ...]  # List of (length, count) tuples
    value_counts: dict[str, int]
    count: int

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            length_range=tuple(resource["lengthRange"]),
            distinct_count=resource["distinctCount"],
            length_histogram=tuple((length, count) for length, count in zip(*resource["lengthHistogram"])),
            value_counts={value: count for value, count in zip(*resource["valueCounts"])},
            count=resource["count"],
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        length_histogram: list[list[int]] = [[], []]
        for length, count in self.length_histogram:
            length_histogram[0].append(length)
            length_histogram[1].append(count)
        value_counts: list[list[str | int]] = [[], []]
        for value, count in self.value_counts.items():
            value_counts[0].append(value)
            value_counts[1].append(count)

        return {
            "lengthRange" if camel_case else "length_range": list(self.length_range),
            "distinctCount" if camel_case else "distinct_count": self.distinct_count,
            "lengthHistogram" if camel_case else "length_histogram": length_histogram,
            "valueCounts" if camel_case else "value_counts": value_counts,
            "count": self.count,
        }


@dataclass
class NumberProfile(CogniteObject):
    value_range: tuple[float, float]
    distinct_count: int
    value_counts: dict[str, int]
    histogram: tuple[tuple[float, int], ...]
    count: int
    mean: float
    std: float
    median: float

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            value_range=tuple(resource["valueRange"]),
            distinct_count=resource["distinctCount"],
            value_counts=resource["valueCounts"],
            histogram=tuple((value, count) for value, count in zip(*resource["histogram"])),
            count=resource["count"],
            mean=resource["mean"],
            std=resource["std"],
            median=resource["median"],
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        histogram: list[list[float | int]] = [[], []]
        for value, count in self.histogram:
            histogram[0].append(value)
            histogram[1].append(count)
        return {
            "valueRange" if camel_case else "value_range": list(self.value_range),
            "distinctCount" if camel_case else "distinct_count": self.distinct_count,
            "valueCounts" if camel_case else "value_counts": self.value_counts,
            "histogram": histogram,
            "count": self.count,
            "mean": self.mean,
            "std": self.std,
            "median": self.median,
        }


@dataclass
class BooleanProfile(CogniteObject):
    count: int
    true_count: int

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            count=resource["count"],
            true_count=resource["trueCount"],
        )


@dataclass
class VectorProfile(CogniteObject):
    count: int
    length_histogram: tuple[tuple[int, int], ...]  # List of (length, count) tuples
    length_range: tuple[int, int]

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            count=resource["count"],
            length_histogram=tuple((length, count) for length, count in zip(*resource["lengthHistogram"])),
            length_range=tuple(resource["lengthRange"]),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        length_histogram: list[list[int]] = [[], []]
        for length, count in self.length_histogram:
            length_histogram[0].append(length)
            length_histogram[1].append(count)
        return {
            "count": self.count,
            "lengthHistogram" if camel_case else "length_histogram": length_histogram,
            "lengthRange" if camel_case else "length_range": list(self.length_range),
        }


@dataclass
class ObjectProfile(CogniteObject):
    count: int
    key_count_histogram: tuple[tuple[int, int], ...]  # List of (key_count, count) tuples
    key_count_range: tuple[int, int]

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            count=resource["count"],
            key_count_histogram=tuple((key_count, count) for key_count, count in zip(*resource["keyCountHistogram"])),
            key_count_range=tuple(resource["keyCountRange"]),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        key_count_histogram: list[list[int]] = [[], []]
        for key_count, count in self.key_count_histogram:
            key_count_histogram[0].append(key_count)
            key_count_histogram[1].append(count)
        return {
            "count": self.count,
            "keyCountHistogram" if camel_case else "key_count_histogram": key_count_histogram,
            "keyCountRange" if camel_case else "key_count_range": list(self.key_count_range),
        }


@dataclass
class StringProfileColumn(RawProfileColumn):
    _type: ClassVar[str] = "string"
    string: StringProfile

    @classmethod
    def _load_column(cls, resource: dict[str, Any]) -> Self:
        return cls(
            count=resource["count"],
            null_count=resource["nullCount"],
            string=StringProfile._load(resource["string"]),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        data = super().dump(camel_case=camel_case)
        data["string"] = self.string.dump(camel_case=camel_case)
        return data


@dataclass
class NumberProfileColumn(RawProfileColumn):
    _type: ClassVar[str] = "number"
    number: NumberProfile

    @classmethod
    def _load_column(cls, resource: dict[str, Any]) -> Self:
        return cls(
            count=resource["count"],
            null_count=resource["nullCount"],
            number=NumberProfile._load(resource["number"]),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        data = super().dump(camel_case=camel_case)
        data["number"] = self.number.dump(camel_case=camel_case)
        return data


@dataclass
class BooleanProfileColumn(RawProfileColumn):
    _type: ClassVar[str] = "boolean"
    boolean: BooleanProfile

    @classmethod
    def _load_column(cls, resource: dict[str, Any]) -> Self:
        return cls(
            count=resource["count"],
            null_count=resource["nullCount"],
            boolean=BooleanProfile._load(resource["boolean"]),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        data = super().dump(camel_case=camel_case)
        data["boolean"] = self.boolean.dump(camel_case=camel_case)
        return data


@dataclass
class VectorProfileColumn(RawProfileColumn):
    _type: ClassVar[str] = "vector"
    vector: VectorProfile

    @classmethod
    def _load_column(cls, resource: dict[str, Any]) -> Self:
        return cls(
            count=resource["count"],
            null_count=resource["nullCount"],
            vector=VectorProfile._load(resource["vector"]),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        data = super().dump(camel_case=camel_case)
        data["vector"] = self.vector.dump(camel_case=camel_case)
        return data


@dataclass
class ObjectProfileColumn(RawProfileColumn):
    _type: ClassVar[str] = "object"
    object: ObjectProfile

    @classmethod
    def _load_column(cls, resource: dict[str, Any]) -> Self:
        return cls(
            count=resource["count"],
            null_count=resource["nullCount"],
            object=ObjectProfile._load(resource["object"]),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        data = super().dump(camel_case=camel_case)
        data["object"] = self.object.dump(camel_case=camel_case)
        return data


@dataclass
class UnknownTypeProfileColumn(RawProfileColumn):
    data: dict[str, object]

    @classmethod
    def _load_column(cls, resource: dict[str, Any]) -> Self:
        return cls(
            count=resource["count"],
            null_count=resource["nullCount"],
            data={k: v for k, v in resource.items() if k not in {"count", "nullCount"}},
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        data = super().dump(camel_case=camel_case)
        data.update(self.data)
        return data


class RawProfileColumns(dict, Mapping[str, RawProfileColumn], CogniteObject):
    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        columns = {name: RawProfileColumn._load(column_data) for name, column_data in resource.items()}
        return cls(columns)

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {name: column.dump(camel_case=camel_case) for name, column in self.items()}

    # The methods are overloads to provide type hints for the methods.
    def items(self) -> ItemsView[str, RawProfileColumn]:  # type: ignore[override]
        return super().items()

    def keys(self) -> KeysView[str]:  # type: ignore[override]
        return super().keys()

    def values(self) -> ValuesView[RawProfileColumn]:  # type: ignore[override]
        return super().values()

    def __iter__(self) -> Iterator[str]:
        yield from super().__iter__()

    def __getitem__(self, column_name: str) -> RawProfileColumn:
        return super().__getitem__(column_name)


class RawProfileResults(CogniteResource):
    """The results of a raw profile operation used with the /profiler/raw endpoint.

    This endpoint is undocumented, and thus it is hard to know exactly what each field means,
    so the following is based on the experimentation with the response of the endpoint.

    Args:
        row_count: The number of rows in the profile.
        columns: The columns in the profile, each column is a RawProfileColumn.
        is_complete: Whether the profiling operation is complete or not. If not complete, the results may be partial.

    """

    def __init__(
        self,
        row_count: int,
        columns: RawProfileColumns,
        is_complete: bool,
    ) -> None:
        self.row_count = row_count
        self.columns = columns
        self.is_complete = is_complete

    @property
    def column_count(self) -> int:
        return len(self.columns)

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            row_count=resource["rowCount"],
            columns=RawProfileColumns._load(resource["columns"]),
            is_complete=resource["isComplete"],
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {
            "rowCount" if camel_case else "row_count": self.row_count,
            "columns": self.columns.dump(camel_case=camel_case),
            "isComplete" if camel_case else "is_complete": self.is_complete,
        }


_RAW_PROFILE_COLUMN_BY_TYPE: MappingProxyType[str, type[RawProfileColumn]] = MappingProxyType(
    # MyPy fails to understand that all subclasses of RawProfileColumn are concrete classes.
    {c._type: c for c in RawProfileColumn.__subclasses__() if not issubclass(c, UnknownTypeProfileColumn)}  # type: ignore[type-abstract]
)



from pathlib import Path

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.commands import DownloadCommand
from cognite_toolkit._cdf_tk.constants import DATA_MANIFEST_STEM
from cognite_toolkit._cdf_tk.storageio import AssetIO, RawIO
from cognite_toolkit._cdf_tk.storageio.selectors import DataSetSelector, RawTableSelector, SelectedTable
from cognite_toolkit._cdf_tk.utils.file import read_yaml_file
from cognite_toolkit._cdf_tk.utils.fileio import CSVReader, NDJsonReader
from tests.test_integration.conftest import HierarchyMinimal
from tests.test_integration.constants import TIMESERIES_COUNT, TIMESERIES_TABLE


class TestDownloadCommand:
    def test_download_raw_table(self, toolkit_client: ToolkitClient, aggregator_raw_db: str, tmp_path: Path) -> None:
        cmd = DownloadCommand(silent=True, skip_tracking=True)
        table = RawTableSelector(table=SelectedTable(db_name=aggregator_raw_db, table_name=TIMESERIES_TABLE))
        cmd.download(
            [table],
            RawIO(toolkit_client),
            output_dir=tmp_path,
            verbose=False,
            file_format=".ndjson",
            compression="none",
            limit=None,
        )
        downloaded_files = list(tmp_path.rglob("*.ndjson"))
        assert len(downloaded_files) == 1, "Expected exactly one file to be downloaded."
        chunks = list(NDJsonReader(downloaded_files[0]).read_chunks())
        assert len(chunks) == TIMESERIES_COUNT, f"Expected {TIMESERIES_COUNT} chunks, got {len(chunks)}."
        config_files = [file for file in tmp_path.rglob("*.yaml") if not file.stem.endswith(DATA_MANIFEST_STEM)]
        assert len(config_files) == 2, (
            f"Expected 2 configuration files (table and database), but found {len(config_files)}."
        )
        table_file = next((file for file in config_files if file.stem.endswith("Table")), None)
        assert table_file is not None, "Table configuration file not found."
        dumped = read_yaml_file(table_file, "dict")
        assert dumped == table.dump()["table"]
        database_file = next((file for file in config_files if file.stem.endswith("Database")), None)
        assert database_file is not None, f"Database configuration file not found. Found config files: {config_files}"
        dumped_db = read_yaml_file(database_file, "dict")
        assert dumped_db == {"dbName": table.table.db_name}

    def test_download_assets(
        self, toolkit_client: ToolkitClient, migration_hierarchy_minimal: HierarchyMinimal, tmp_path: Path
    ) -> None:
        cmd = DownloadCommand(silent=True, skip_tracking=True)
        selected = DataSetSelector(kind="Assets", data_set_external_id=migration_hierarchy_minimal.dataset.external_id)
        cmd.download(
            [selected],
            AssetIO(toolkit_client),
            output_dir=tmp_path,
            verbose=False,
            file_format=".csv",
            compression="none",
            limit=None,
        )
        downloaded_files = list(tmp_path.rglob("*.csv"))
        assert len(downloaded_files) == 1, "Expected exactly one file to be downloaded."
        chunks = list(CSVReader(downloaded_files[0]).read_chunks_unprocessed())
        assert len(chunks) == 2, f"Expected 2 assets, got {len(chunks)}."
        columns = set(chunks[0].keys())
        expected_aggregate_columns = {"depth", "childCount", "path"}
        assert expected_aggregate_columns <= columns, (
            f"Expected hierarchy columns missing in downloaded data: {columns}"
        )



import json
import sys
import warnings
from dataclasses import dataclass
from typing import Any, Literal

from cognite.client import CogniteClient
from cognite.client.data_classes._base import CogniteObject
from cognite.client.data_classes.data_modeling import DirectRelationReference
from cognite.client.data_classes.data_modeling.ids import ViewId
from cognite.client.data_classes.data_modeling.instances import (
    PropertyOptions,
    TypedNode,
    TypedNodeApply,
)

from cognite_toolkit._cdf_tk.constants import COGNITE_MIGRATION_SPACE
from cognite_toolkit._cdf_tk.tk_warnings import IgnoredValueWarning
from cognite_toolkit._cdf_tk.utils.useful_types import AssetCentricType

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


@dataclass(frozen=True)
class AssetCentricId(CogniteObject):
    resource_type: AssetCentricType
    id_: int

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        """Load an AssetCentricId from a dictionary."""
        return cls(
            resource_type=resource["resourceType"],
            id_=resource["id"],
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        """Dump the AssetCentricId to a dictionary."""
        return {
            "resourceType" if camel_case else "resource_type": self.resource_type,
            "id" if camel_case else "id_": self.id_,
        }

    def __str__(self) -> str:
        """Return a string representation of the AssetCentricId."""
        return f"{self.resource_type}(id={self.id_})"


class _InstanceSourceProperties:
    resource_type = PropertyOptions("resourceType")
    id_ = PropertyOptions("id")
    data_set_id = PropertyOptions("dataSetId")
    classic_external_id = PropertyOptions("classicExternalId")
    preferred_consumer_view_id = PropertyOptions("preferredConsumerViewId")
    ingestion_view = PropertyOptions("ingestionView")

    @classmethod
    def get_source(cls) -> ViewId:
        return ViewId("cognite_migration", "InstanceSource", "v1")


class InstanceSource(_InstanceSourceProperties, TypedNode):
    """This represents the reading format of instance source.

    It is used to when data is read from CDF.

    The source of the instance in asset-centric resources.

    Args:
        space: The space where the node is located.
        external_id: The external id of the instance source.
        version (int): DMS version.
        last_updated_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970,
            Coordinated Universal Time (UTC), minus leap seconds.
        created_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970,
            Coordinated Universal Time (UTC), minus leap seconds.
        resource_type: The resource type field.
        id_: The id field.
        data_set_id: The data set id field.
        classic_external_id: The classic external id field.
        preferred_consumer_view_id: The preferred consumer view id field.
        ingestion_view: The ingestion view field.
        type: Direct relation pointing to the type node.
        deleted_time: The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time
            (UTC), minus leap seconds. Timestamp when the instance was soft deleted. Note that deleted instances
            are filtered out of query results, but present in sync results
    """

    def __init__(
        self,
        space: str,
        external_id: str,
        version: int,
        last_updated_time: int,
        created_time: int,
        *,
        resource_type: AssetCentricType,
        id_: int,
        data_set_id: int | None = None,
        classic_external_id: str | None = None,
        preferred_consumer_view_id: ViewId | None = None,
        ingestion_view: DirectRelationReference | None = None,
        type: DirectRelationReference | None = None,
        deleted_time: int | None = None,
    ) -> None:
        TypedNode.__init__(self, space, external_id, version, last_updated_time, created_time, deleted_time, type)
        self.resource_type = resource_type
        self.id_ = id_
        self.data_set_id = data_set_id
        self.classic_external_id = classic_external_id
        self.preferred_consumer_view_id = preferred_consumer_view_id
        self.ingestion_view = DirectRelationReference.load(ingestion_view) if ingestion_view else None

    @classmethod
    def _load_properties(cls, resource: dict[str, Any]) -> dict[str, Any]:
        if "preferredConsumerViewId" in resource:
            preferred_consumer_view_id = resource.pop("preferredConsumerViewId")
            try:
                resource["preferredConsumerViewId"] = ViewId.load(preferred_consumer_view_id)
            except (TypeError, KeyError) as e:
                warnings.warn(
                    IgnoredValueWarning(
                        name="InstanceSource.preferredConsumerViewId",
                        value=json.dumps(preferred_consumer_view_id),
                        reason=f"Invalid ViewId format expected 'space', 'externalId', 'version': {e!s}",
                    )
                )
        return super()._load_properties(resource)

    def as_asset_centric_id(self) -> AssetCentricId:
        """Return the AssetCentricId representation of the mapping."""
        return AssetCentricId(
            resource_type=self.resource_type,
            id_=self.id_,
        )

    def consumer_view(self) -> ViewId:
        if self.preferred_consumer_view_id:
            return self.preferred_consumer_view_id
        if self.resource_type == "sequence":
            raise ValueError(f"Missing consumer view for sequence {self.external_id}.")
        # Default consumer view for asset-centric resources
        external_id = {
            "asset": "CogniteAsset",
            "event": "CogniteActivity",
            "file": "CogniteFile",
            "timeseries": "CogniteTimeSeries",
        }[self.resource_type]
        return ViewId("cdf_cdm", external_id, "v1")

    def as_direct_relation_reference(self) -> DirectRelationReference:
        return DirectRelationReference(space=self.space, external_id=self.external_id)


class _ResourceViewMapping:
    resource_type = PropertyOptions("resourceType")
    view_id = PropertyOptions("viewId")
    property_mapping = PropertyOptions("propertyMapping")

    @classmethod
    def get_source(cls) -> ViewId:
        return ViewId("cognite_migration", "ResourceViewMapping", "v1")


class ResourceViewMappingApply(_ResourceViewMapping, TypedNodeApply):
    """This represents the writing format of view source.

    It is used to when data is written to CDF.

    The source of the view in asset-centric resources.

    Args:
        external_id: The external id of the view source.
        resource_type: The resource type field.
        view_id: The view id field.
        property_mapping: The mapping of asset-centric properties to data model properties.
        existing_version: Fail the ingestion request if the node's version is greater than or equal to this value.
            If no existingVersion is specified, the ingestion will always overwrite any existing data for the node
            (for the specified container or node). If existingVersion is set to 0, the upsert will behave as an insert,
            so it will fail the bulk if the item already exists. If skipOnVersionConflict is set on the ingestion
            request, then the item will be skipped instead of failing the ingestion request.
        type: Direct relation pointing to the type node.
    """

    def __init__(
        self,
        external_id: str,
        *,
        resource_type: str,
        view_id: ViewId,
        property_mapping: dict[str, str],
        existing_version: int | None = None,
        type: DirectRelationReference | tuple[str, str] | None = None,
    ) -> None:
        TypedNodeApply.__init__(self, COGNITE_MIGRATION_SPACE, external_id, existing_version, type)
        self.resource_type = resource_type
        self.view_id = view_id
        self.property_mapping = property_mapping

    def dump(self, camel_case: bool = True, context: Literal["api", "local"] = "api") -> dict[str, Any]:
        """Dumps the object to a dictionary.

        Args:
            camel_case: Whether to use camel case or not.
            context: If 'api', the output is for the API and will match the Node API schema. If 'local', the output is
                for a YAML file and all properties are  on the same level as the node properties. See below

        Example:
            >>> node = ResourceViewMappingApply(
            ...    external_id="myMapping",
            ...    resource_type="asset",
            ...    view_id=ViewId("cdf_cdm", "CogniteAsset", "v1"),
            ...    property_mapping={"name": "name"},
            ... )
            >>> node.dump(camel_case=True, context="api")
            {
                "space": "cognite_migration",
                "externalId": "myMapping",
                "sources": [
                    {
                        "source": {
                            "space": "cognite_migration",
                            "externalId": "ViewSource",
                            "version": "v1",
                            "type": "view"
                            },

                        "properties": {
                            "resourceType": "asset",
                            "viewId": {
                                "space": "cdf_cdm",
                                "externalId": "CogniteAsset",
                                "version": "v1"
                            },
                            "propertyMapping": {
                                "name": "name"
                            }
                        }
                    }
                ]
            }
            >>> node.dump(camel_case=True, context="local")
            {
                "externalId": "myMapping",
                "resourceType": "asset",
                "viewId": {
                    "space": "cdf_cdm",
                    "externalId": "CogniteAsset",
                    "version": "v1"
                },
                "propertyMapping": {
                    "name": "name"
                },
            }

        Returns:
            dict[str, Any]: The dumped dictionary representation of the object.
        """
        output = super().dump(camel_case)
        source = output["sources"][0]
        properties = source["properties"]
        properties["viewId"] = self.view_id.dump(camel_case=camel_case, include_type=context == "api")

        if context == "local":
            for key in ("space", "sources", "instanceType"):
                output.pop(key, None)
            output.update(properties)
        return output

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        base_props = cls._load_base_properties(resource)
        properties = cls._load_properties(resource)
        if "viewId" in resource:
            properties["view_id"] = ViewId.load(resource["viewId"])

        return cls(**base_props, **properties)


class ResourceViewMapping(_ResourceViewMapping, TypedNode):
    """This represents the reading format of view source.

    It is used to when data is read from CDF.

    The source of the view in asset-centric resources.

    Args:
        external_id: The external id of the view source.
        version (int): DMS version.
        last_updated_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970,
            Coordinated Universal Time (UTC), minus leap seconds.
        created_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970,
            Coordinated Universal Time (UTC), minus leap seconds.
        resource_type: The resource type field.
        view_id: The view id field.
        property_mapping: The mapping field.
        type: Direct relation pointing to the type node.
        deleted_time: The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time
            (UTC), minus leap seconds. Timestamp when the instance was soft deleted. Note that deleted instances
            are filtered out of query results, but present in sync results
    """

    def __init__(
        self,
        external_id: str,
        version: int,
        last_updated_time: int,
        created_time: int,
        *,
        resource_type: str,
        view_id: ViewId,
        property_mapping: dict[str, str],
        type: DirectRelationReference | None = None,
        deleted_time: int | None = None,
    ) -> None:
        TypedNode.__init__(
            self, COGNITE_MIGRATION_SPACE, external_id, version, last_updated_time, created_time, deleted_time, type
        )
        self.resource_type = resource_type
        self.view_id = view_id
        self.property_mapping = property_mapping

    def as_write(self) -> ResourceViewMappingApply:
        return ResourceViewMappingApply(
            self.external_id,
            resource_type=self.resource_type,
            view_id=self.view_id,
            property_mapping=self.property_mapping,
            existing_version=self.version,
            type=self.type,
        )

    @classmethod
    def _load_properties(cls, resource: dict[str, Any]) -> dict[str, Any]:
        if "viewId" in resource:
            view_id = resource.pop("viewId")
            try:
                resource["viewId"] = ViewId.load(view_id)
            except (TypeError, KeyError) as e:
                raise ValueError(f"Invalid viewId format. Expected 'space', 'externalId', 'version'. Error: {e!s}")
        return super()._load_properties(resource)

    def _dump_properties(self) -> dict[str, Any]:
        """Dump the properties of the ViewSourceApply."""
        return {
            "resourceType": self.resource_type,
            "viewId": self.view_id.dump(),
            "propertyMapping": self.property_mapping,
        }


class CreatedSourceSystem(TypedNode):
    """This represents the reading format of instance source.

    It is used to when data is read from CDF.

    The source of the instance in asset-centric resources.

    Args:
        space: The space where the node is located.
        external_id: The external id of the instance source.
        version (int): DMS version.
        last_updated_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970,
            Coordinated Universal Time (UTC), minus leap seconds.
        created_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970,
            Coordinated Universal Time (UTC), minus leap seconds.
        source: The source string from the asset-centric Asset, Event, or FileMetadata source property.
        type: Direct relation pointing to the type node.
        deleted_time: The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time
            (UTC), minus leap seconds. Timestamp when the instance was soft deleted. Note that deleted instances
            are filtered out of query results, but present in sync results
    """

    def __init__(
        self,
        space: str,
        external_id: str,
        version: int,
        last_updated_time: int,
        created_time: int,
        *,
        source: str,
        type: DirectRelationReference | None = None,
        deleted_time: int | None = None,
    ) -> None:
        TypedNode.__init__(self, space, external_id, version, last_updated_time, created_time, deleted_time, type)
        self.source = source

    @classmethod
    def get_source(cls) -> ViewId:
        return ViewId("cognite_migration", "CreatedSourceSystem", "v1")

    def as_direct_relation_reference(self) -> DirectRelationReference:
        return DirectRelationReference(space=self.space, external_id=self.external_id)


class SpaceSource(TypedNode):
    """This represents the reading format of space source.

    It is used to when data is read from CDF.

    The source dataSet of an instance space

    Args:
        space: The space where the node is located.
        external_id: The external id of the node.
        version (int): DMS version.
        last_updated_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970,
            Coordinated Universal Time (UTC), minus leap seconds.
        created_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970,
            Coordinated Universal Time (UTC), minus leap seconds.
        instance_space (str): The instance space name.
        data_set_id: The data set id associated with the instance space.
        data_set_external_id: The classic external id associated with the instance space.
        type: Direct relation pointing to the type node.
        deleted_time: The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time
            (UTC), minus leap seconds. Timestamp when the instance was soft deleted. Note that deleted instances
            are filtered out of query results, but present in sync results
    """

    instance_space = PropertyOptions("instanceSpace")
    data_set_id = PropertyOptions("dataSetId")
    data_set_external_id = PropertyOptions("dataSetExternalId")

    def __init__(
        self,
        space: str,
        external_id: str,
        version: int,
        last_updated_time: int,
        created_time: int,
        *,
        instance_space: str,
        data_set_id: int,
        data_set_external_id: str | None = None,
        type: DirectRelationReference | None = None,
        deleted_time: int | None = None,
    ) -> None:
        TypedNode.__init__(self, space, external_id, version, last_updated_time, created_time, deleted_time, type)
        self.instance_space = instance_space
        self.data_set_id = data_set_id
        self.data_set_external_id = data_set_external_id

    @classmethod
    def get_source(cls) -> ViewId:
        return ViewId("cognite_migration", "SpaceSource", "v1")



from collections.abc import Callable, Iterable
from pathlib import Path
from typing import Any

from cognite_toolkit._cdf_tk.builders import Builder
from cognite_toolkit._cdf_tk.constants import BUILD_FOLDER_ENCODING
from cognite_toolkit._cdf_tk.cruds import TransformationCRUD
from cognite_toolkit._cdf_tk.data_classes import (
    BuildDestinationFile,
    BuildSourceFile,
    ModuleLocation,
    SourceLocation,
)
from cognite_toolkit._cdf_tk.exceptions import ToolkitYAMLFormatError
from cognite_toolkit._cdf_tk.tk_warnings import HighSeverityWarning, ToolkitWarning
from cognite_toolkit._cdf_tk.utils import safe_write


class TransformationBuilder(Builder):
    _resource_folder = TransformationCRUD.folder_name

    def build(
        self, source_files: list[BuildSourceFile], module: ModuleLocation, console: Callable[[str], None] | None = None
    ) -> Iterable[BuildDestinationFile | list[ToolkitWarning]]:
        query_files = {
            source_file.source.path: source_file
            for source_file in source_files
            if source_file.source.path.suffix == ".sql"
        }

        for source_file in source_files:
            loaded = source_file.loaded
            if loaded is None:
                # Not a YAML file
                continue
            loader, warning = self._get_loader(source_file.source.path)
            if loader is None:
                if warning is not None:
                    yield [warning]
                continue

            destination_path = self._create_destination_path(source_file.source.path, loader.kind)

            extra_sources: list[SourceLocation] | None = None
            if loader is TransformationCRUD:
                extra_sources = self._add_query(loaded, source_file, query_files, destination_path)

            destination = BuildDestinationFile(
                path=destination_path,
                loaded=loaded,
                loader=loader,
                source=source_file.source,
                extra_sources=extra_sources,
            )
            yield destination

    def load_extra_field(self, extra: str) -> tuple[str, Any]:
        return "query", extra

    def _add_query(
        self,
        loaded: dict[str, Any] | list[dict[str, Any]],
        source_file: BuildSourceFile,
        query_files: dict[Path, BuildSourceFile],
        transformation_destination_path: Path,
    ) -> list[SourceLocation]:
        loaded_list = loaded if isinstance(loaded, list) else [loaded]
        extra_sources: list[SourceLocation] = []
        for entry in loaded_list:
            try:
                external_id = TransformationCRUD.get_id(entry)
            except KeyError:
                # This will be validated later
                continue
            filepath = source_file.source.path
            query_file = self._get_query_file(filepath, external_id, query_files)

            if "query" in entry and query_file is not None:
                raise ToolkitYAMLFormatError(
                    f"query property is ambiguously defined in both the yaml file and a separate file named {query_file}\n"
                    f"Please remove one of the definitions, either the query property in {filepath} or the file {query_file}",
                )
            elif "query" not in entry and query_file is None:
                warning = HighSeverityWarning(
                    f"query property or is missing in {filepath.as_posix()!r}. It can be inline or a separate file named {filepath.stem}.sql or {external_id}.sql",
                )
                if self.warn:
                    self.warn(warning)
                else:
                    warning.print_warning()
            elif query_file is not None:
                destination_path = self._create_destination_path(query_file.source.path, "Query")
                safe_write(destination_path, query_file.content, encoding=BUILD_FOLDER_ENCODING)
                relative = destination_path.relative_to(transformation_destination_path.parent)
                entry["queryFile"] = relative.as_posix()
                extra_sources.append(query_file.source)

        return extra_sources

    @staticmethod
    def _get_query_file(
        source_file: Path, transformation_external_id: str | None, query_files: dict[Path, BuildSourceFile]
    ) -> BuildSourceFile | None:
        query_file = source_file.parent / f"{source_file.stem}.sql"
        if query_file in query_files:
            return query_files[query_file]
        if transformation_external_id:
            query_file = source_file.parent / f"{transformation_external_id}.sql"
            if query_file in query_files:
                return query_files[query_file]
        return None



"""Data class for command tracking information."""

from typing import Any

from pydantic import BaseModel, Field


class CommandTrackingInfo(BaseModel):
    """Structured tracking information for CLI commands.

    This model provides type-safe tracking information that can be collected
    during command execution and sent to Mixpanel for analytics.

    Attributes:
        project: The CDF project name.
        cluster: The CDF cluster name.
        module_ids: List of module IDs that were deployed or built.
        package_ids: List of package IDs that were deployed or built.
        installed_module_ids: List of module IDs that were installed.
        installed_package_ids: List of package IDs that were installed.
        downloaded_library_ids: List of library IDs that were downloaded.
        downloaded_package_ids: List of package IDs that were downloaded.
        downloaded_module_ids: List of module IDs that were downloaded.
    """

    project: str | None = Field(default=None)
    cluster: str | None = Field(default=None)
    module_ids: set[str] = Field(default_factory=set, alias="moduleIds")
    package_ids: set[str] = Field(default_factory=set, alias="packageIds")
    installed_module_ids: set[str] = Field(default_factory=set, alias="installedModuleIds")
    installed_package_ids: set[str] = Field(default_factory=set, alias="installedPackageIds")
    downloaded_library_ids: set[str] = Field(default_factory=set, alias="downloadedLibraryIds")
    downloaded_package_ids: set[str] = Field(default_factory=set, alias="downloadedPackageIds")
    downloaded_module_ids: set[str] = Field(default_factory=set, alias="downloadedModuleIds")

    def to_dict(self) -> dict[str, Any]:
        """Convert the tracking info to a dictionary for Mixpanel.

        Returns:
            A dictionary with camelCase keys matching Mixpanel's expected format.
            Default values are excluded.
        """
        return self.model_dump(mode="json", by_alias=True, exclude_defaults=True)



from cognite.client.data_classes.data_modeling import Space

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.data_classes.infield import InfieldLocationConfig
from tests.test_integration.constants import RUN_UNIQUE_ID


class TestInfieldConfig:
    def test_create_retrieve_delete(self, toolkit_client: ToolkitClient, toolkit_space: Space) -> None:
        config = InfieldLocationConfig.model_validate(
            {
                "space": toolkit_space.space,
                "externalId": f"test_crud_infield_config_{RUN_UNIQUE_ID}",
                "rootLocationExternalId": "test_crud_infield_config",
                "dataExplorationConfig": {"observations": {"enabled": True}},
            }
        )

        try:
            created_list = toolkit_client.infield.config.apply([config])
            assert len(created_list) == 2, (
                "Expected 2 configs to be created (data exploration config and infield location config)"
            )
            created = created_list[0]
            assert created.as_id() == config.as_id()

            retrieved_configs = toolkit_client.infield.config.retrieve([config.as_id()])
            assert len(retrieved_configs) == 1
            assert retrieved_configs[0].dump() == config.dump()

            deleted = toolkit_client.infield.config.delete([config])
            assert len(deleted) == 2, (
                "Expected 2 configs to be deleted (data exploration config and infield location config)"
            )
            retrieved_configs = toolkit_client.infield.config.retrieve([config.as_id()])
            assert len(retrieved_configs) == 0
        finally:
            toolkit_client.data_modeling.instances.delete(
                [
                    (config.space, config.external_id),
                    (config.data_exploration_config.space, config.data_exploration_config.external_id),
                ]
            )
